# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gi.repository import GObject

import os
from typing import Optional
from urllib.parse import quote


class Attachment(GObject.Object):
    """A note attachment."""

    def __init__(self) -> None:
        super().__init__()
        self.__note_id = -1
        self.__note_remote_id: Optional[int] = None
        self.__path = ""

    @GObject.Property(type=int, default=-1)
    def note_id(self) -> int:
        return self.__note_id

    @note_id.setter
    def set_note_id(self, value: int) -> None:
        self.__note_id = value

    @GObject.Property(type=int, default=-1)
    def note_remote_id(self) -> Optional[int]:
        return self.__note_remote_id

    @note_remote_id.setter
    def set_note_remote_id(self, value: int) -> None:
        self.__note_remote_id = value

    @GObject.Property(type=str, default="")
    def filename(self) -> str:
        return os.path.basename(self.__path)

    @GObject.Property(type=str, default="")
    def path(self) -> str:
        return self.__path

    @path.setter
    def set_path(self, value: str) -> None:
        self.__path = value

    @GObject.Property(type=str, default="")
    def path_quoted(self) -> str:
        return quote(self.path)

    def duplicate(self) -> Attachment:
        """Duplicate attachment.

        Only the shallow fields are copied; the data on disk is not duplicated.

        :return: A copy of the attachment
        :rtype:
        """
        ret = Attachment()
        ret.note_id = self.note_id
        ret.note_remote_id = self.note_remote_id
        ret.path = self.path
        return ret



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gi.repository import GLib, GObject

import glob
import json
import logging
import os
import shutil
from typing import Optional

from iotas.backup_storage import BackupStorage
from iotas.note import Note, DirtyFields
from iotas.text_utils import sanitise_path


class BackupManager(GObject.Object):

    DATA_SUBPATH = "iotas"
    PRIMARY_DIRNAME = "backup"
    ARCHIVE_DIRNAME = "backup-previous"
    METADATA_EXTENSION = "iota"
    EXPORT_SCHEMA_VERSION = "1.0"
    MAXIMUM_CONTENT_SIZE_MB = 100

    def __init__(
        self, db: BackupStorage, iotas_version: str, storage_dir: Optional[str] = None
    ) -> None:
        super().__init__()
        self.__db = db
        self.__iotas_version = iotas_version

        if storage_dir:
            working_directory = storage_dir
        else:
            working_directory = os.path.join(GLib.get_user_data_dir(), self.DATA_SUBPATH)
        self.__primary_backup_path = os.path.join(working_directory, self.PRIMARY_DIRNAME)
        self.__archive_backup_path = os.path.join(working_directory, self.ARCHIVE_DIRNAME)

        self.__file_extension = "md"
        self.__sync_configured = False

    def create_backup(self, sync_configured: bool, file_extension: str) -> bool:
        """Create a backup.

        :param bool sync_configured: Whether sync is configured
        :param str file_extension: File extension to use for notes
        :return: Whether successful
        :rtype: bool
        """
        self.__sync_configured = sync_configured
        self.__file_extension = file_extension

        if not os.path.exists(self.__primary_backup_path):
            try:
                os.mkdir(self.__primary_backup_path)
            except OSError as e:
                logging.error(
                    f"Failed to create backup directory at {self.__primary_backup_path}: {e}"
                )
                return False

        if not self.__switch_old_backup_to_archive():
            logging.error("Failed to move previous backup to archive path")
            return False

        success = True
        notes = self.__db.get_all_notes(load_content=True)
        for note in notes:
            success, content_filename = self.__write_note_content(note)
            if not success:
                break

            meta_filename = f"{content_filename}.{self.METADATA_EXTENSION}"
            success = self.__write_note_metadata(note, meta_filename)
            if not success:
                break

        if success:
            logging.info(f"Backup created at {self.__primary_backup_path}")
            if sync_configured:
                logging.warning(
                    "Backup restoration isn't possible with Nextcloud Notes sync configured"
                )
        else:
            logging.error("Backup failed")
        return success

    def restore_backup(self, sync_configured: bool) -> bool:
        """Restore a backup.

        :param bool sync_configured: Whether sync is configured
        :return: Whether successful
        :rtype: bool
        """

        # A fairly light touch has been taken here. Presuming that existing backups are
        # unmodified/sane restoring to an empty session should be fairly straight forward.
        #
        # Some work has been done towards merging into an existing collection and restoring to a
        # remotely connect session but as they haven't been thoroughly tested yet they're not
        # exposed.

        self.__sync_configured = sync_configured

        if not os.path.exists(self.__primary_backup_path):
            logging.error(f"No backup exists at {self.__primary_backup_path}")
            return False

        file_list = glob.glob(
            os.path.join(self.__primary_backup_path, f"*.{self.METADATA_EXTENSION}")
        )
        if len(file_list) == 0:
            logging.error(f"No backup exists at {self.__primary_backup_path}")
            return False

        # Notify and quit if attempting to merge a backup into existing notes
        if self.__db.get_all_notes_count() > 0:
            logging.error("Backup restoration can only be run when there are no existing notes")
            return False

        # Notify and quit if attempting to restore a backup into a synced session
        if sync_configured:
            logging.error("Backup restoration isn't possible with Nextcloud Notes sync configured")
            return False

        # Load all notes first, to avoid dealing with transactions
        notes = []
        success = True
        for meta_full_path in file_list:
            note = self.__load_note(meta_full_path)
            if note is None:
                success = False
                break
            notes.append(note)

        if not success:
            logging.error("Backup restoration failed")
            return False

        for note in notes:
            self.__restore_note(note)

        logging.info("Backup restoration completed")
        return True

    def __write_note_metadata(self, note: Note, meta_filename: str) -> bool:
        content = {
            "SchemaVersion": self.EXPORT_SCHEMA_VERSION,
            "IotasVersion": self.__iotas_version,
            "Title": note.title,
            "Category": note.category,
            "LastModified": note.last_modified,
            "Favourite": note.favourite,
            "Dirty": note.dirty,
            "LocallyDeleted": note.locally_deleted,
            "RemoteId": note.remote_id,
            "ETag": note.etag,
        }

        filename = os.path.join(self.__primary_backup_path, meta_filename)
        success = False
        try:
            with open(filename, "w") as f:
                f.write(json.dumps(content, indent=2))
                success = True
        except OSError as e:
            logging.warning(f"Failed to write metadata to {filename}: {e}")
        return success

    def __write_note_content(self, note: Note) -> tuple[bool, str]:
        title = sanitise_path(note.title)

        content_filename = f"{title}.{self.__file_extension}"
        filename = os.path.join(self.__primary_backup_path, content_filename)
        success = False
        try:
            with open(filename, "w") as f:
                f.write(note.content)
                success = True
        except OSError as e:
            logging.warning(f"Failed to write metadata to {filename}: {e}")
        return (success, content_filename)

    def __restore_note(self, note: Note) -> None:
        # Check for existing remote id
        remote_id_note = None
        if self.__sync_configured and note.remote_id > -1:
            remote_id_note = self.__db.get_note_by_remote_id(note.remote_id)
        if remote_id_note is not None:
            self.__restore_note_with_matching_remote_id(note, remote_id_note)
        else:
            existing_note = self.__db.get_note_by_title(note.title)
            if existing_note is not None:
                self.__restore_note_with_matching_title(note, existing_note)
            else:
                # Create new
                logging.debug(f"Creating {note.title}")
                self.__db.add_note(note)

    def __restore_note_with_matching_remote_id(self, note: Note, remote_id_note: Note) -> None:
        if note.identical_to(remote_id_note):
            logging.debug(f'Skipping restoration of existing identical note "{note.title}"')
        else:
            # Translators: Description, prefixes note title on backup restoration clash
            duplicate_reason = _("RESTORATION REMOTE ID CLASH")
            self.__db.create_duplicate_note(note, duplicate_reason)
            logging.info(f'Duplicating note "{note.title}" due to matching remote id')

    def __restore_note_with_matching_title(self, note: Note, existing_note: Note):
        if note.identical_to(existing_note):
            logging.debug(f'Skipping restoration of existing identical note "{note.title}"')
        elif note.content == existing_note.content:
            if note.last_modified > existing_note.last_modified:
                # Update metadata if last modified newer and content matches
                self.__update_note_metadata(existing_note, note)
                logging.info(f'Updating metadata for note "{note.title}"')
            else:
                msg = 'Skipping note "{}" with matching title, contents and a newer timestamp'
                logging.info(msg.format(note.title))
        else:
            # Duplicate note
            # Translators: Description, prefixes note title on backup restoration clash
            duplicate_reason = _("RESTORATION TITLE CLASH")
            self.__db.create_duplicate_note(note, duplicate_reason)
            logging.info(
                f'Duplicating note "{note.title}" due to matching title but different content'
            )

    def __update_note_metadata(self, dest: Note, source: Note) -> None:
        # Update metadata if last modified newer and content matches
        changed_fields = DirtyFields()
        if dest.favourite != source.favourite:
            dest.favourite = source.favourite
            changed_fields.favourite = True
        if dest.last_modified != source.last_modified:
            dest.last_modified = source.last_modified
            changed_fields.last_modified = True
        if dest.category != source.category:
            dest.category = source.category
            changed_fields.category = True
        if dest.locally_deleted != source.locally_deleted:
            dest.locally_deleted = source.locally_deleted
            changed_fields.locally_deleted = True
        if dest.etag != source.etag:
            dest.etag = source.etag
            changed_fields.etag = True
        if dest.remote_id != source.remote_id:
            dest.remote_id = source.remote_id
            changed_fields.remote_id = True
        self.__db.persist_note_selective(dest, changed_fields)

    def __switch_old_backup_to_archive(self) -> bool:
        file_list = glob.glob(
            os.path.join(self.__primary_backup_path, f"*.{self.METADATA_EXTENSION}")
        )
        if len(file_list) == 0:
            return True

        archive_path = self.__archive_backup_path
        if os.path.exists(archive_path):
            try:
                shutil.rmtree(archive_path)
            except OSError as e:
                logging.error(f"Failed to remove backup archive directory at {archive_path}: {e}")
                return False

        try:
            os.mkdir(archive_path)
        except OSError as e:
            logging.error(f"Failed to create backup archive directory at {archive_path}: {e}")
            return False

        for meta_file in file_list:
            content_file = ".".join(meta_file.split(".")[:-1])
            if os.path.exists(content_file):
                try:
                    shutil.move(content_file, archive_path)
                except OSError as e:
                    logging.error(f"Failed to move {content_file} into {archive_path}: {e}")
                    return False
            try:
                shutil.move(meta_file, archive_path)
            except OSError as e:
                logging.error(f"Failed to move {meta_file} into {archive_path}: {e}")
                return False

        return True

    def __load_note(self, meta_full_path: str) -> Optional[Note]:
        meta_filename = os.path.basename(meta_full_path)
        content_full_path = ".".join(meta_full_path.split(".")[:-1])
        content_filename = os.path.basename(content_full_path)
        if not os.path.exists(content_full_path):
            logging.warning(f'Skipping "{content_filename}" due to missing content')
            return None

        # Skip restoring massive files
        if os.path.getsize(content_full_path) > self.MAXIMUM_CONTENT_SIZE_MB * 1000 * 1000:
            logging.error(
                f'Bailing "{content_filename}" due to exceeding {self.MAXIMUM_CONTENT_SIZE_MB}MB'
            )
            return None

        try:
            with open(content_full_path, "r") as f:
                content = f.read()
        except OSError as e:
            logging.error(f"Failed to read note content from {content_full_path}: {e}")
            return None

        try:
            with open(meta_full_path, "r") as f:
                meta_str = f.read()
                meta = json.loads(meta_str)
        except (OSError, json.decoder.JSONDecodeError) as e:
            logging.error(f"Failed to read note metadata from {meta_filename}: {e}")
            return None

        note = Note.from_backup(meta, content, self.__sync_configured)
        if note is None:
            logging.error(f'Failed to parse note metadata from "{meta_filename}"')
            return None

        return note

    @staticmethod
    def get_default_primary_path() -> str:
        return os.path.join(
            GLib.get_user_data_dir(), BackupManager.DATA_SUBPATH, BackupManager.PRIMARY_DIRNAME
        )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
import gi

gi.require_version("GtkSource", "5")

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Graphene, Gtk, GtkSource, Pango

from dataclasses import dataclass
import logging
from typing import Any, Optional
import webbrowser

from iotas.attachment import Attachment
from iotas.category_header_bar import CategoryHeaderBar
from iotas.category_treeview_list_store import CategoryTreeViewListStore
from iotas.config_manager import ConfigManager, HeaderBarVisibility
import iotas.const as const
from iotas.editor_rename_header_bar import EditorRenameHeaderBar
from iotas.editor_search_header_bar import EditorSearchHeaderBar
from iotas.editor_text_view import EditorTextView
from iotas.export_dialog import ExportDialog
from iotas.focus_mode_helper import FocusModeHelper
from iotas.font_size_selector import FontSizeSelector
from iotas.formatting_header_bar import FormattingHeaderBar
from iotas.note import Note
from iotas.outline_dialog import OutlineDialog
from iotas.outline_generator import OutlineHeading, OutlineGenerator
from iotas.render_search_header_bar import RenderSearchHeaderBar
from iotas.sync_manager import SyncManager
from iotas.text_utils import (
    parse_any_url_at_iter,
    parse_markdown_automatic_link,
    parse_markdown_inline_link,
)
from iotas.theme_selector import ThemeSelector
from iotas.ui_utils import add_mouse_button_accel, is_likely_mobile_device


@dataclass
class EditorSessionDetails:
    note: Note
    scroll_position: float
    cursor_position: int
    direct_switching: Optional[bool] = None


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/editor.ui")
class Editor(Adw.Bin):
    __gtype_name__ = "Editor"
    __gsignals__ = {
        "note-modified": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        "note-deleted": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        "category-changed": (GObject.SignalFlags.RUN_FIRST, None, (Note, str)),
        "exit": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    LINE_LENGTH_STEP = 50
    TOAST_DURATION = 3.0
    HEADERBAR_HIDE_DELAY = 2000
    MARGIN_BELOW_CURSOR = 22

    _top_bar_revealer: Gtk.Revealer = Gtk.Template.Child()
    _bottom_bar_revealer: Gtk.Revealer = Gtk.Template.Child()

    _title_label: Gtk.Label = Gtk.Template.Child()
    _is_dirty: Gtk.Label = Gtk.Template.Child()
    _headerbar_stack: Gtk.Stack = Gtk.Template.Child()
    _main_header_bar: Adw.HeaderBar = Gtk.Template.Child()
    _rename_header_bar: EditorRenameHeaderBar = Gtk.Template.Child()
    _search_header_bar: EditorSearchHeaderBar = Gtk.Template.Child()
    _category_header_bar: CategoryHeaderBar = Gtk.Template.Child()
    _formatting_header_bar: FormattingHeaderBar = Gtk.Template.Child()
    _render_search_header_bar: RenderSearchHeaderBar = Gtk.Template.Child()

    _menu_button: Gtk.MenuButton = Gtk.Template.Child()
    _title_button: Gtk.Button = Gtk.Template.Child()
    _render_edit_stack: Gtk.Stack = Gtk.Template.Child()
    _render_edit_button_stack: Gtk.Stack = Gtk.Template.Child()
    _render_button: Gtk.Button = Gtk.Template.Child()
    _edit_button: Gtk.Button = Gtk.Template.Child()
    _sourceview: EditorTextView = Gtk.Template.Child()
    _editor_scrolledwin: Gtk.ScrolledWindow = Gtk.Template.Child()
    _read_only_image: Gtk.Image = Gtk.Template.Child()
    _render_loading: Adw.StatusPage = Gtk.Template.Child()
    _toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self.__active = False
        self.__note: Optional[Note] = None
        self.__note_had_edit = False
        self.__updating_from_remote = False
        self.__sync_authenticated = False
        self.__buffer_handlers: list[int] = []
        self.__current_note_handlers: list[int] = []
        # Any types to accommodate for lazy loading
        self.__render_view: Optional[Any] = None
        self.__html_generator: Optional[Any] = None
        self.__focus_mode_action: Gio.SimpleAction
        self.__cancel_action: Gio.SimpleAction
        self.__previous_sessions: list[EditorSessionDetails] = []
        self.__outline_dialog: Optional[OutlineDialog] = None
        self.__hiding_bars_timeout: Optional[int] = None
        self.__pointer_position: Graphene.Point = Graphene.Point.zero()

        self.__outline_generator = OutlineGenerator()

        self.__config_manager = ConfigManager.get_default()

        self._menu_button.get_popover().add_child(ThemeSelector(), "theme")
        self.__font_size_selector = FontSizeSelector()
        self._menu_button.get_popover().add_child(self.__font_size_selector, "fontsize")

        self._category_header_bar.connect("categories-changed", self.__on_category_changed)
        self._category_header_bar.connect("abort", self.__on_abort_category_change)

        self._search_header_bar.connect(
            "resumed", lambda _o: self.__enter_search(resuming=True, for_replace=False)
        )
        self._search_header_bar.connect(
            "open-for-replace", lambda _o: self.__enter_search(resuming=False, for_replace=True)
        )
        self._search_header_bar.bind_property(
            "replace-visible",
            self._sourceview,
            "accommodate-double-header-bar",
            GObject.BindingFlags.SYNC_CREATE,
        )

        self._render_search_header_bar.connect(
            "resumed", lambda _o: self.__enter_search(resuming=True, for_replace=False)
        )

        self._rename_header_bar.connect("renamed", self.__on_title_renamed)
        self._rename_header_bar.connect("cancelled", self.__on_rename_cancelled)

        self._formatting_header_bar.connect(
            "height-change", lambda _o: self.__update_toolbar_underlay_margins()
        )

        self._render_edit_button_stack.set_visible(self.__config_manager.markdown_render_enabled)

        self.__categories_model = None

        style_manager = Adw.StyleManager.get_default()
        style_manager.connect("notify::dark", lambda _o, _v: self.__update_scheme_and_dark_style())
        style_manager.connect(
            "notify::high-contrast", lambda _o, _v: self.__update_scheme_and_dark_style()
        )

        self.__config_manager.connect_changed(
            ConfigManager.MARKDOWN_DETECT_SYNTAX,
            self.__on_detect_syntax_changed,
        )
        self.__config_manager.connect_changed(
            ConfigManager.EDITOR_THEME,
            self.__on_theme_changed,
        )
        self.__config_manager.connect_changed(
            ConfigManager.MARKDOWN_RENDER,
            self.__on_render_enabled_changed,
        )
        self.__config_manager.connect_changed(
            ConfigManager.MARKDOWN_KEEP_WEBKIT_PROCESS,
            self.__on_keep_webkit_changed,
        )
        self.__config_manager.connect_changed(
            ConfigManager.MARKDOWN_TEX_SUPPORT,
            self.__on_tex_support_changed,
        )
        self.__config_manager.connect_changed(
            ConfigManager.LINE_LENGTH,
            self.__refresh_line_length_from_setting,
        )

        self.__push_font_updates()
        style_manager.connect(
            "notify::monospace-font-name", lambda _o, _v: self.__push_font_updates()
        )
        style_manager.connect(
            "notify::document-font-name", lambda _o, _v: self.__push_font_updates()
        )
        self.__config_manager.connect_changed(
            ConfigManager.USE_MONOSPACE_FONT,
            self.__push_font_updates,
        )
        self.__config_manager.connect_changed(
            ConfigManager.MARKDOWN_USE_MONOSPACE_FONT,
            self.__push_font_updates,
        )

        self.__config_manager.connect_changed(
            ConfigManager.EDITOR_FORMATTING_BAR_VISIBILTY,
            self.__update_for_formatting_bar_setting_change,
        )
        self.__cache_formatting_bar_visibility()

        self.__config_manager.connect_changed(
            ConfigManager.EDITOR_HEADER_BAR_VISIBILTY,
            self.__update_for_header_bar_setting_change,
        )
        self.__cache_header_bar_visibility()

        self.__focus_mode_helper = FocusModeHelper(self._sourceview.get_buffer())

    def setup(
        self,
        model: CategoryTreeViewListStore,
        sync_manager: SyncManager,
    ) -> None:
        """Perform initial setup.

        :param CategoryTreeViewListStore model: Tree model
        :param SyncManager sync_manager: Sync manager
        """
        self.__setup_actions()
        self._category_header_bar.setup(model)
        self.__sync_manager = sync_manager

        self._search_header_bar.setup(self._sourceview)
        self._render_search_header_bar.setup()
        self.__font_size_selector.setup()
        self.__refresh_line_length_from_setting()
        self._formatting_header_bar.setup(self._sourceview)

        window = self.get_root()
        window.connect("notify::fullscreened", self.__on_window_fullscreened)
        self._sourceview.setup()

    def init_note(self, note: Note, restore_session: Optional[EditorSessionDetails]) -> None:
        """Initialise editing session.

        :param Note note: Note to edit
        :param EditorSessionDetails restore_session: Session to restore, optional
        """
        self.current_note = note
        self.note_had_edit = False
        buffer = self._sourceview.get_buffer()
        for handler_id in self.__buffer_handlers:
            buffer.disconnect(handler_id)
        self._sourceview.set_visible(True)
        self._sourceview.spellchecker_enabled = False
        self._formatting_header_bar.prepare()
        buffer.set_text(note.content)

        def delayed_enable_spellchecker():
            self._sourceview.spellchecker_enabled = True

        if self.__config_manager.spelling_enabled:
            GLib.idle_add(delayed_enable_spellchecker)

        self.__buffer_handlers = []
        handler_id = buffer.connect("changed", lambda _o: self.__on_buffer_changed())
        self.__buffer_handlers.append(handler_id)

        self.__setup_syntax_detection()
        self.__setup_note_signals()
        self.__update_title()
        self.__update_dirty()

        self._read_only_image.set_visible(note.read_only)
        self._sourceview.set_editable(not note.read_only)
        self.__update_readonly_actions()

        # Handle case of leaflet swiped back to index while renaming or searching
        if self._headerbar_stack.get_visible_child() in (
            self._rename_header_bar,
            self._search_header_bar,
            self._category_header_bar,
            self._render_search_header_bar,
        ):
            self._headerbar_stack.set_visible_child(self._main_header_bar)

        if self._render_edit_stack.get_visible_child() == self.__render_view:
            self.__toggle_render()
        else:
            # This only sets the widget visibility; there's logic below which determines whether,
            # based on configuration, the bottom bar is shown
            self._formatting_header_bar.set_visible(True)

        # By popping this note off our previous sessions it allows returning from the index to
        # switching between two notes
        self.__flush_note_from_previous_sessions(note)

        if restore_session:
            buffer = self._sourceview.get_buffer()
            cursor_iter = buffer.get_iter_at_offset(restore_session.cursor_position)
            scroll_position = restore_session.scroll_position
            GLib.idle_add(self._editor_scrolledwin.get_vadjustment().set_value, scroll_position)
        else:
            # Set scroll to top
            cursor_iter = buffer.get_start_iter()
            scroll_position = 0
        self._editor_scrolledwin.get_vadjustment().set_value(scroll_position)

        buffer.place_cursor(cursor_iter)

        self.__update_scheme_and_dark_style()

        if not (restore_session and restore_session.direct_switching):
            self._top_bar_revealer.set_reveal_child(True)
            if (
                self.__config_manager.markdown_detect_syntax
                and self.__formatting_bar_visibility_setting != HeaderBarVisibility.DISABLED
            ):
                self._bottom_bar_revealer.set_reveal_child(True)
            elif self._bottom_bar_revealer.get_reveal_child():
                self._bottom_bar_revealer.set_reveal_child(False)
            GLib.idle_add(self.__update_toolbar_underlay_margins)

            self.__check_and_hide_bars_after_delay()

        if self.__focus_mode_action.get_state():
            self.__focus_mode_helper.active = True

        # Handle initialising in view mode
        if (
            self.__config_manager.markdown_render_enabled
            and self.__config_manager.markdown_default_to_render
            and note.content.strip() != ""
        ):
            # Prevent a flash of the editor while loading rendered view
            self._sourceview.set_visible(False)
            # The idle avoids locked high CPU usage after first render
            GLib.idle_add(self.__toggle_render)

        if is_likely_mobile_device():
            if self.__config_manager.editor_formatting_bar_visibility in (
                HeaderBarVisibility.AUTO_HIDE,
                HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY,
            ):
                logging.warning("Auto hiding currently provides a poor experience on mobile")

        if restore_session:
            self.__restore_session_location(restore_session)

    def close_note(self) -> None:
        """Close the current editing note."""
        if not self.__note:
            return

        self.__close_any_header_bar(after_exit=True)

        self.__store_session_details()

        if self.__outline_dialog:
            self.__outline_dialog.close()

        if self.current_note.title_is_top_line:
            self.current_note.title_is_top_line = False

        for handler in self.__current_note_handlers:
            self.__note.disconnect(handler)
        self.__current_note_handlers = []

        self.__focus_mode_helper.active = False
        self._bottom_bar_revealer.set_visible(True)
        self._search_header_bar.disable_actions()
        self._render_search_header_bar.disable_actions()
        if self.__hiding_bars_timeout:
            GLib.source_remove(self.__hiding_bars_timeout)
            self.__hiding_bars_timeout = None

        self.current_note = None

        # If cleaning up WebKit process after each render do that now on closing the note.
        # Don't do this for other cases as the switch is visible as the navigation page
        # animates across.
        if not self.__config_manager.markdown_keep_webkit_process:
            if self._render_edit_stack.get_visible_child() == self.__render_view:
                self._sourceview.set_visible(True)
                self.__toggle_render()

    def cancel(self, checking_after_exit: bool = False) -> None:
        """Close any open header bar or exit the editor."""
        if not self.__close_any_header_bar(after_exit=False):
            self.emit("exit")

    def set_sync_authenticated(self, authenticated: bool) -> None:
        """Set that an authenticated sync session exists.

        :param bool authenticated: Whether authenticated
        """
        self.__sync_authenticated = authenticated

    def focus_textview_if_editing(self) -> None:
        """Grab the focus to the TextView."""
        # Don't focus if showing render view
        if self._render_edit_stack.get_visible_child() != self._editor_scrolledwin:
            return

        # Don't focus if search etc toolbar is active
        if (
            self._top_bar_revealer.get_reveal_child()
            and self._headerbar_stack.get_visible_child() != self._main_header_bar
        ):
            return

        self._sourceview.grab_focus()

    def update_for_dialog_visibility(self, visible: bool) -> None:
        """Update if dialogs visible, disabling actions.

        :param bool visible: Whether a dialog is visible
        """
        if self.active:
            self.__cancel_action.set_enabled(not visible)

    def get_previous_sessions(self) -> list[EditorSessionDetails]:
        """Get previous detail for previous sessions.

        :return: List of session details
        :rtype: list[EditorSessionDetails]
        """
        return list(self.__previous_sessions)

    def show_downloaded_attachment_in_render(self, attachment: Attachment) -> None:
        """Update render to display newly available attachment.

        :param Attachment attachment: Attachment to show
        """
        if self.__note is None or attachment.note_id != self.__note.id:
            return
        if self._render_edit_stack.get_visible_child() != self.__render_view:
            return
        logging.debug(f"Attachment '{attachment.filename}' downloaded")
        self.__render_view.show_attached_image(attachment)

    # Return type intentionally skipped to allow for lazy loading WebKit
    def ensure_webkit_initialised(self) -> Any:
        """Ensure WebKit is initialised."""
        if self.__render_view is None:
            self.__init_webkit()
        return self.__render_view

    def ensure_html_generator_initialised(self) -> Any:
        """Ensure the HTML generator is initialised."""
        if self.__html_generator is None:
            self.__initialise_html_generator()
            self.__push_font_updates()

        return self.__html_generator

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value
        self.__enable_actions(value)
        if value and self.__config_manager.markdown_detect_syntax:
            self._formatting_header_bar.active = True
        else:
            self._formatting_header_bar.active = False

    @GObject.Property(type=Note, default=None)
    def current_note(self) -> Optional[Note]:
        return self.__note

    @current_note.setter
    def set_current_note(self, value: Note) -> None:
        self.__note = value

    @GObject.Property(type=bool, default=False)
    def editing(self) -> bool:
        """Whether editing.

        :returns: Whether editing
        :rtype: bool
        """
        # Using the editor having an active note to determine if
        return self.__note is not None

    @GObject.Property(type=bool, default=False)
    def note_had_edit(self) -> bool:
        """Get whether note was modified during session.

        Determines whether excerpt will later be updated.

        :returns: Whether modified
        :rtype: bool
        """
        return self.__note_had_edit

    @note_had_edit.setter
    def set_note_had_edit(self, value: bool) -> None:
        """Set whether note was modified during session."""
        self.__note_had_edit = value

    @Gtk.Template.Callback()
    def _on_mouse_back_button_click(
        self, _gesture: Gtk.GestureClick, _n_press: int, _x: float, _y: float
    ) -> None:
        self.cancel()

    @Gtk.Template.Callback()
    def _on_editor_single_click(
        self, gesture: Gtk.GestureSingle, _seq: Optional[Gdk.EventSequence] = None
    ) -> None:
        # If the click is in the visible top row of the buffer where the show-triggering motion area
        # will be don't hide otherwise we immediately hide and re-show, which is glitchy.
        if not self.__click_in_top_margin(gesture):
            self.__check_and_hide_header_bar()

        if not self.__click_in_bottom_margin(gesture):
            self.__check_and_hide_formatting_bar()

        event = gesture.get_current_event()
        if event is None:
            return
        state = event.get_modifier_state()
        if state & Gdk.ModifierType.CONTROL_MASK and state & Gdk.ModifierType.BUTTON1_MASK:
            _interpreted, x, y = gesture.get_point(None)
            self.__check_and_handle_link_click(x, y)

    @Gtk.Template.Callback()
    def _on_top_margin_motion(
        self, _controller: Gtk.EventControllerMotion, x: float, y: float
    ) -> None:
        self.__process_top_margin_input(x, y)

    @Gtk.Template.Callback()
    def _on_top_margin_click(
        self, gesture: Gtk.GestureClick, n_press: int, x: float, y: float
    ) -> None:
        self.__process_top_margin_input(x, y)

    @Gtk.Template.Callback()
    def _on_bottom_margin_motion(
        self, _controller: Gtk.EventControllerMotion, x: float, y: float
    ) -> None:
        self.__process_bottom_margin_input(x, y)

    @Gtk.Template.Callback()
    def _on_bottom_margin_click(
        self, gesture: Gtk.GestureClick, n_press: int, x: float, y: float
    ) -> None:
        self.__process_bottom_margin_input(x, y)

    @Gtk.Template.Callback()
    def _on_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        # Prevent some keys from resulting in the bars re-hiding
        if keyval not in (
            Gdk.KEY_Tab,
            Gdk.KEY_ISO_Left_Tab,
            Gdk.KEY_Alt_L,
            Gdk.KEY_Alt_R,
            Gdk.KEY_Meta_L,
            Gdk.KEY_Meta_R,
            Gdk.KEY_Control_L,
            Gdk.KEY_Control_R,
            Gdk.KEY_Shift_L,
            Gdk.KEY_Shift_R,
        ):
            self.__check_and_hide_header_bar()
            self.__check_and_hide_formatting_bar()

        return Gdk.EVENT_PROPAGATE

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("rename")
        action.connect("activate", lambda _o, _p: self.__apply_rename())
        action_group.add_action(action)
        app.set_accels_for_action("editor.rename", ["F2"])

        action = Gio.SimpleAction.new("edit-category")
        action.connect("activate", lambda _o, _p: self.__edit_category())
        action_group.add_action(action)
        app.set_accels_for_action("editor.edit-category", ["<Control>e"])

        action = Gio.SimpleAction.new("enter-search")
        action.connect("activate", lambda _o, _p: self.__on_enter_search())
        action_group.add_action(action)
        app.set_accels_for_action("editor.enter-search", ["<Control>f"])

        action = Gio.SimpleAction.new("toggle-render")
        action.connect("activate", lambda _o, _p: self.__toggle_render())
        action_group.add_action(action)
        app.set_accels_for_action("editor.toggle-render", ["<Control>d"])

        action = Gio.SimpleAction.new("go-back-or-cancel")
        action.connect("activate", lambda _o, _p: self.cancel())
        action_group.add_action(action)
        self.__cancel_action = action
        app.set_accels_for_action("editor.go-back-or-cancel", ["Escape"])

        action = Gio.SimpleAction.new("clear-category")
        action.connect("activate", lambda _o, _p: self.__on_clear_category())
        action_group.add_action(action)
        app.set_accels_for_action("editor.clear-category", ["<Shift>Delete"])

        action = Gio.SimpleAction.new("create-new-note")
        action.connect("activate", lambda _o, _v: self.__on_create_new_note())
        action_group.add_action(action)
        app.set_accels_for_action("editor.create-new-note", ["<Control>n"])

        action = Gio.SimpleAction.new("export-note")
        action.connect("activate", lambda _o, _p: self.__on_export_note())
        action_group.add_action(action)
        app.set_accels_for_action("editor.export-note", ["<Control><Shift>s"])

        action = Gio.SimpleAction.new("show-outline")
        action.connect("activate", lambda _o, _p: self.__show_outline())
        action_group.add_action(action)
        app.set_accels_for_action("editor.show-outline", ["<Control>j"])

        action = Gio.SimpleAction.new("delete-note")
        action.connect("activate", lambda _o, _p: self.emit("note-deleted", self.__note))
        action_group.add_action(action)

        action = Gio.SimpleAction.new("increase-line-length")
        action.connect("activate", lambda _o, _p: self.__on_increase_line_length())
        action_group.add_action(action)
        app.set_accels_for_action("editor.increase-line-length", ["<Control>Up"])

        action = Gio.SimpleAction.new("decrease-line-length")
        action.connect("activate", lambda _o, _p: self.__on_decrease_line_length())
        action_group.add_action(action)
        app.set_accels_for_action("editor.decrease-line-length", ["<Control>Down"])

        action = Gio.SimpleAction.new("toggle-line-length-limit")
        action.connect("activate", lambda _o, _p: self.__on_line_length_limit_toggle())
        action_group.add_action(action)
        app.set_accels_for_action("editor.toggle-line-length-limit", ["<Shift><Control>l"])

        action = Gio.SimpleAction.new("increase-font-size")
        action.connect("activate", lambda _o, _v: self.__font_size_selector.increase())
        action_group.add_action(action)
        app.set_accels_for_action("editor.increase-font-size", ["<Control>plus", "<Control>equal"])

        action = Gio.SimpleAction.new("decrease-font-size")
        action.connect("activate", lambda _o, _v: self.__font_size_selector.decrease())
        action_group.add_action(action)
        app.set_accels_for_action(
            "editor.decrease-font-size", ["<Control>minus", "<Control>underscore"]
        )

        action = Gio.SimpleAction.new("reset-font-size")
        action.connect("activate", lambda _o, _v: self.__font_size_selector.reset())
        action_group.add_action(action)
        app.set_accels_for_action("editor.reset-font-size", ["<Control>0"])

        action = Gio.SimpleAction.new_stateful("focus-mode", None, GLib.Variant("b", False))
        action.connect("change-state", self.__on_focus_mode_toggle)
        action_group.add_action(action)
        app.set_accels_for_action("editor.focus-mode", ["<Control><Shift>f"])
        self.__focus_mode_action = action

        action = Gio.SimpleAction.new("show-menu")
        action.connect("activate", lambda _o, _p: self._menu_button.popup())
        action_group.add_action(action)
        app.set_accels_for_action("editor.show-menu", ["F10"])

        action = Gio.SimpleAction.new("focus-editor")
        action.connect("activate", lambda _o, _p: self.__focus_editor())
        action_group.add_action(action)
        app.set_accels_for_action("editor.focus-editor", ["<Alt>e"])

        action = Gio.SimpleAction.new("focus-header-bar")
        action.connect("activate", lambda _o, _p: self.__focus_header_bar())
        action_group.add_action(action)
        app.set_accels_for_action("editor.focus-header-bar", ["<Alt>h"])

        action = Gio.SimpleAction.new("focus-formatting-bar")
        action.connect("activate", lambda _o, _p: self.__focus_formatting_bar())
        action_group.add_action(action)
        app.set_accels_for_action("editor.focus-formatting-bar", ["<Alt>f"])

        app.get_active_window().insert_action_group("font-size-selector", action_group)
        self.__action_group = action_group

        self.__action_group = action_group
        app.get_active_window().insert_action_group("editor", action_group)

    def __enable_actions(self, enabled: bool) -> None:
        """Toggle whether editor actions are enabled.

        :param bool enabled: New value
        """
        actions = self.__action_group.list_actions()
        for action in actions:
            self.__action_group.lookup_action(action).set_enabled(enabled)
        # Disable specific actions for read-only
        if enabled and self.__note is not None and self.__note.read_only:
            self.__update_readonly_actions()

    def __setup_syntax_detection(self) -> None:
        buffer = self._sourceview.get_buffer()
        if self.__config_manager.markdown_detect_syntax:
            language = GtkSource.LanguageManager.get_default().get_language("iotas-markdown")
            buffer.set_language(language)
            buffer.set_highlight_syntax(True)
        else:
            buffer.set_language(None)
            buffer.set_highlight_syntax(False)

    def __on_checkbox_toggled(self, _obj: GObject.Object, line: int, new_value: bool) -> None:
        buffer = self._sourceview.get_buffer()

        def get_iter_inside_checkbox(line):
            success, line_start = buffer.get_iter_at_line(line)
            if not success:
                return None

            line_end = line_start.copy()
            line_end.forward_to_line_end()
            match = line_start.forward_search("[", Gtk.TextSearchFlags.VISIBLE_ONLY, line_end)
            if match is None:
                logging.warn(f"Couldn't find checkbox on line {line}")
                return None

            _, match_end = match
            return match_end

        match_end = get_iter_inside_checkbox(line)
        if match_end is None:
            return

        buffer.begin_user_action()
        buffer.insert(match_end, "x" if new_value else " ")

        match_end = get_iter_inside_checkbox(line)
        if match_end is None:
            buffer.end_user_action()
            return
        match_end.forward_char()
        delete_end = match_end.copy()
        delete_end.forward_char()
        buffer.delete(match_end, delete_end)
        buffer.end_user_action()

    def __on_webview_loaded(self, _obj: GObject.Object) -> None:
        assert self.__render_view
        if self.__render_view.exporting:
            self.__render_view.exporting = False
        else:
            self._render_edit_button_stack.set_visible_child(self._edit_button)
            if not self.__render_view.get_visible():
                self.__render_view.set_visible(True)
            self._render_edit_stack.set_visible_child(self.__render_view)

    def __on_enter_search(self) -> None:
        if self._search_header_bar.active:
            self._search_header_bar.refocus_search_and_select()
        else:
            self.__enter_search(resuming=False, for_replace=False)

    def __on_title_renamed(self, _obj: GObject.Object, new_title: str) -> None:
        assert self.__note
        if self.__note.title != new_title:
            self.__note.title = new_title
            if self.__note.title_is_top_line:
                self.__note.title_is_top_line = False
            self.__note.flag_changed()
            self.note_had_edit = True
            # Always called on UI thread
            self.emit("note-modified", self.__note)
        self._headerbar_stack.set_visible_child(self._main_header_bar)
        self.__check_and_hide_bars_after_delay()
        self._sourceview.grab_focus()

    def __on_rename_cancelled(self, _obj: GObject.Object) -> None:
        self._headerbar_stack.set_visible_child(self._main_header_bar)
        self.__check_and_hide_bars_after_delay()
        self._sourceview.grab_focus()

    def __on_clear_category(self) -> None:
        if not self._top_bar_revealer.get_reveal_child():
            return
        if self._headerbar_stack.get_visible_child() == self._category_header_bar:
            self._category_header_bar.clear_and_apply()

    def __on_export_note(self) -> None:
        assert self.__note
        if self.__html_generator is None:
            self.__initialise_html_generator()
            self.__push_font_updates()

        dialog = ExportDialog(
            [self.__note],
            self.__html_generator,
            self.__sync_manager.download_attachment if self.__sync_manager.authenticated else None,
            self.ensure_webkit_initialised,
            note_content_populator=None,
        )
        dialog.present(self)

    def __on_detect_syntax_changed(self) -> None:
        if self.active:
            self.__setup_syntax_detection()
            if self.__config_manager.markdown_detect_syntax:
                if self.__formatting_bar_visibility_setting != HeaderBarVisibility.DISABLED:
                    self.__show_formatting_bar()
                self._formatting_header_bar.active = True
            else:
                self.__hide_formatting_bar()
                self._formatting_header_bar.active = False

    def __on_theme_changed(self) -> None:
        if self.active and self.__config_manager.markdown_detect_syntax:
            self.__update_scheme_and_dark_style()

    def __on_render_enabled_changed(self) -> None:
        enabled = self.__config_manager.markdown_render_enabled
        self._render_edit_button_stack.set_visible(enabled)
        if (
            not enabled
            and self.active
            and self._render_edit_stack.get_visible_child() == self.__render_view
        ):
            self.__toggle_render(force=True)

    def __on_keep_webkit_changed(self) -> None:
        keep = self.__config_manager.markdown_keep_webkit_process
        if not keep and (
            not self.active
            or self._render_edit_stack.get_visible_child() == self._editor_scrolledwin
        ):
            if self.__render_view is not None:
                logging.debug("Calling terminate for WebKit process")
                self.__render_view.terminate_web_process()

    def __on_tex_support_changed(self) -> None:
        assert self.__note
        if self.active and self._render_edit_stack.get_visible_child() == self.__render_view:
            self.__render_view.render_retaining_scroll(self.__note, None)

    def __on_increase_line_length(self) -> None:
        setting_max = self.__config_manager.line_length_max
        current_length = self.__config_manager.line_length
        if current_length == setting_max:
            logging.info("Line length limit already disabled")
            return
        else:
            new_length = current_length + self.LINE_LENGTH_STEP
        if new_length < self.get_width() - 2 * self._sourceview.STANDARD_X_MARGIN:
            self.__config_manager.line_length = new_length
            logging.info(f"Line length increased to {new_length}px")
        else:
            self.__config_manager.line_length = setting_max
            self.__notify_line_length_limit_disabled()

    def __on_decrease_line_length(self) -> None:
        setting_max = self.__config_manager.line_length_max
        current_length = self.__config_manager.line_length
        new_length = current_length - self.LINE_LENGTH_STEP
        if current_length == setting_max:
            new_length = min(self.get_width(), self.__config_manager.default_line_length)
            self.__config_manager.line_length = new_length
            self.__notify_line_length_limit_enabled(new_length)
        elif new_length >= self.LINE_LENGTH_STEP:
            self.__config_manager.line_length = new_length
            logging.info(f"Line length decreased to {new_length}px")

    def __on_line_length_limit_toggle(self) -> None:
        setting_maximum = self.__config_manager.line_length_max
        if self.__config_manager.line_length == setting_maximum:
            new_length = self.__config_manager.default_line_length
            self.__config_manager.line_length = new_length
            self.__notify_line_length_limit_enabled(new_length)
        else:
            self.__config_manager.line_length = setting_maximum
            self.__notify_line_length_limit_disabled()

    def __on_window_fullscreened(self, window: Gtk.Window, _param: GObject.ParamSpec) -> None:
        if not self.active:
            return

        if self._render_edit_stack.get_visible_child() == self.__render_view:
            self.__render_view.update_style()

        if self.__header_bar_visibility_setting == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
            if window.is_fullscreen():
                self.__hide_header_bar()
            else:
                self.__show_header_bar()

        if self.__config_manager.markdown_detect_syntax:
            setting = self.__formatting_bar_visibility_setting
            if setting == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
                if window.is_fullscreen():
                    self.__hide_formatting_bar()
                else:
                    self.__show_formatting_bar()

        self.__update_toolbar_underlay_margins()

    def __on_render_single_click(self) -> None:
        self.__check_and_hide_header_bar()

    def __on_render_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if self._headerbar_stack.get_visible_child() != self._main_header_bar:
            return Gdk.EVENT_PROPAGATE

        result = self._render_search_header_bar.check_if_starting(controller, keyval, state)
        if result:
            self._render_search_header_bar.enter(self.__render_view, False, type_to_search=True)
            self._headerbar_stack.set_visible_child(self._render_search_header_bar)
            self.__show_header_bar()
            return Gdk.EVENT_STOP
        else:
            return Gdk.EVENT_PROPAGATE

    def __on_category_changed(self, _obj: CategoryHeaderBar) -> None:
        changeset = self._category_header_bar.take_changeset()
        if len(changeset) > 0:
            note, old_category = changeset[0]
            self.emit("category-changed", note, old_category)
        self._headerbar_stack.set_visible_child(self._main_header_bar)
        self.__check_and_hide_bars_after_delay()
        self._sourceview.grab_focus()

    def __on_abort_category_change(self, _obj: CategoryHeaderBar) -> None:
        self._headerbar_stack.set_visible_child(self._main_header_bar)
        self.__check_and_hide_bars_after_delay()
        self._sourceview.grab_focus()

    def __on_focus_mode_toggle(self, action: Gio.SimpleAction, state: bool) -> None:
        action.set_state(state)
        self.__focus_mode_helper.active = state

    def __on_create_new_note(self) -> None:
        self.emit("exit")

        content = ""
        buffer = self._sourceview.get_buffer()
        if buffer.get_has_selection():
            begin, end = buffer.get_selection_bounds()
            content = buffer.get_slice(begin, end, False)
        param = GLib.Variant("s", content)

        self.activate_action("index.create-note-with-content", param)

    def __on_buffer_changed(self) -> None:
        if self.__updating_from_remote:
            return

        # Witnessed when closing for conflict
        if not self.__note:
            return

        self.__note.content = self.__get_content()
        renaming = self._headerbar_stack.get_visible_child() == self._rename_header_bar
        if self.__note.title_is_top_line and not renaming:
            self.__note.update_title_from_top_line()
        self.__note.flag_changed()
        self.note_had_edit = True

        # Maintain a margin between the cursor and the bottom of the window
        self.__check_and_scroll_for_cursor()

        GLib.idle_add(self.emit, "note-modified", self.__note)

    def set_header_bar_height(self, height: int) -> None:
        """Set header bar height for top margin.

        :param int height: Bar height
        """
        self._sourceview.header_bar_height = height

    def __check_and_scroll_for_cursor(self) -> None:
        buffer = self._sourceview.get_buffer()
        insert = buffer.get_iter_at_mark(buffer.get_insert())

        rect = self._sourceview.get_iter_location(insert)
        adj = self._editor_scrolledwin.get_vadjustment()
        lowest_visible = self._sourceview.get_top_margin() + rect.y + rect.height
        visible_below_cursor = adj.get_value() + adj.get_page_size() - lowest_visible
        if visible_below_cursor < self.MARGIN_BELOW_CURSOR:
            scroll_to = lowest_visible + self.MARGIN_BELOW_CURSOR - adj.get_page_size()
            adj.set_value(scroll_to)

    def __notify_line_length_limit_enabled(self, new_length: int) -> None:
        logging.info(f"Line length re-enabled, to {new_length}px")
        # Translators: Description, notification, {0} is a number
        msg = _("Line length now {0}px").format(new_length)
        self.__show_toast(msg)

    def __notify_line_length_limit_disabled(self) -> None:
        logging.info("Line length limit disabled")
        # Translators: Description, notification
        self.__show_toast(_("Line length limit disabled"))

    def __show_toast(self, msg: str) -> None:
        if not self.active:
            logging.debug(f"Not showing toast as editor inactive: {msg}")
            return

        toast = Adw.Toast.new(msg)
        toast.set_priority(Adw.ToastPriority.HIGH)
        toast.set_timeout(self.TOAST_DURATION)

        self._toast_overlay.dismiss_all()
        self._toast_overlay.add_toast(toast)

    def __init_webkit(self) -> None:
        # Lazy import module in an attempt to reduce startup performance hit on devices with
        # render view disabled
        from iotas.markdown_render_view import MarkdownRenderView

        if self.__html_generator is None:
            self.__initialise_html_generator()
            assert self.__html_generator  # mypy

        app = Gio.Application.get_default()
        track_timing = app.debug_session or app.development_mode

        self.__render_view = MarkdownRenderView()
        self.__render_view.connect("checkbox-toggled", self.__on_checkbox_toggled)
        self.__render_view.connect("loaded", self.__on_webview_loaded)
        self.__render_view.setup(self.__html_generator, self.__config_manager, track_timing)

        self._render_edit_stack.add_child(self.__render_view)

        gesture = Gtk.GestureSingle()
        gesture.connect("end", lambda _o, _s: self.__on_render_single_click())
        self.__render_view.add_controller(gesture)

        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__on_render_key_pressed)
        self.__render_view.add_controller(controller)

        self.__push_font_updates()

        self._render_search_header_bar.bind_property(
            "active", self.__render_view, "searching", GObject.BindingFlags.SYNC_CREATE
        )

        # Allow using the mouse back button to return to index, typically button #8
        def callback(gesture: Gtk.GestureClick, _n_press: int, _x: float, _y: float) -> None:
            if gesture.get_current_button() == 8:
                self.cancel()

        add_mouse_button_accel(self.__render_view, callback)

    def __close_any_header_bar(self, after_exit: bool) -> bool:
        closed_headerbar = False
        if self._headerbar_stack.get_visible_child() in (
            self._rename_header_bar,
            self._category_header_bar,
        ):
            self._headerbar_stack.set_visible_child(self._main_header_bar)
            if not after_exit:
                self._sourceview.grab_focus()
                self.__check_and_hide_bars_after_delay()
            closed_headerbar = True
        elif self._headerbar_stack.get_visible_child() == self._search_header_bar:
            self.__exit_search()
            closed_headerbar = True
        elif self._headerbar_stack.get_visible_child() == self._render_search_header_bar:
            self.__exit_render_search()
            closed_headerbar = True

        return closed_headerbar

    def __setup_note_signals(self) -> None:
        assert self.__note
        self.__current_note_handlers = []

        handler_id = self.__note.connect("notify::dirty", lambda _o, _v: self.__update_dirty())
        self.__current_note_handlers.append(handler_id)
        handler_id = self.__note.connect("notify::title", lambda _o, _v: self.__update_title())
        self.__current_note_handlers.append(handler_id)

        handler_id = self.__note.connect("remote-content-update", self.__update_content_from_remote)
        self.__current_note_handlers.append(handler_id)

    def __get_content(self) -> str:
        buffer = self._sourceview.get_buffer()
        start, end = buffer.get_bounds()
        return buffer.get_text(start, end, True)

    def __update_title(self) -> None:
        assert self.__note
        self._title_label.set_label(self.__note.title)

    def __update_dirty(self) -> None:
        assert self.__note
        show_dirty = self.__sync_authenticated and self.__note.dirty and self.__note.title != ""
        self._is_dirty.set_visible(show_dirty)

    def __update_content_from_remote(self, _note: Note) -> None:
        if self.__note and not self.__note.dirty:
            self.set_sensitive(False)
            self.__updating_from_remote = True
            editor_scroll_value = self._editor_scrolledwin.get_vadjustment().get_value()
            buffer = self._sourceview.get_buffer()
            insert_iter = buffer.get_iter_at_mark(buffer.get_insert())
            pre_offset = insert_iter.get_offset()
            buffer.set_text(self.__note.content)

            def delayed_editor_scroll_and_cursor_reset():
                self._editor_scrolledwin.get_vadjustment().set_value(editor_scroll_value)
                if self.__note and len(self.__note.content) > pre_offset:
                    buffer.place_cursor(buffer.get_iter_at_offset(pre_offset))

            GLib.idle_add(delayed_editor_scroll_and_cursor_reset)
            if self._render_edit_stack.get_visible_child() == self.__render_view:
                self.__render_view.render_retaining_scroll(self.__note, export_format=None)
            self.__updating_from_remote = False
            self.set_sensitive(True)

            # Update outline from remote
            if self.__outline_dialog:
                self.__show_outline()

    def __get_scroll_position(self) -> float:
        adj = self._editor_scrolledwin.get_vadjustment()
        value = adj.get_value()
        upper = adj.get_upper()
        page_size = adj.get_page_size()
        if upper > page_size:
            return value / (upper - page_size)
        else:
            return 0.0

    def __set_scroll_position(self, position: float) -> None:
        adj = self._editor_scrolledwin.get_vadjustment()
        upper = adj.get_upper()
        page_size = adj.get_page_size()
        adj.set_value((upper - page_size) * position)

    def __edit_category(self) -> None:
        """Show the category editing."""
        assert self.__note
        if not self._top_bar_revealer.get_reveal_child():
            self.__show_header_bar()
        if self._headerbar_stack.get_visible_child() == self._category_header_bar:
            return
        elif self._headerbar_stack.get_visible_child() == self._search_header_bar:
            self.__exit_search()
        elif self._headerbar_stack.get_visible_child() == self._render_search_header_bar:
            self.__exit_render_search()
        self._headerbar_stack.set_visible_child(self._category_header_bar)
        self._category_header_bar.activate([self.__note])

    def __apply_rename(self) -> None:
        """Show the title rename entry."""
        assert self.__note
        if self.__note.read_only:
            return
        if not self._top_bar_revealer.get_reveal_child():
            self.__show_header_bar()
        if self._headerbar_stack.get_visible_child() == self._rename_header_bar:
            return
        elif self._headerbar_stack.get_visible_child() == self._search_header_bar:
            self.__exit_search()
        elif self._headerbar_stack.get_visible_child() == self._render_search_header_bar:
            self.__exit_render_search()
        self._headerbar_stack.set_visible_child(self._rename_header_bar)
        self._rename_header_bar.enter(self.__note.title)

    def __enter_search(self, resuming: bool, for_replace: bool) -> None:
        """Start searching within note."""
        if self._render_edit_stack.get_visible_child() == self._editor_scrolledwin:
            self._search_header_bar.enter(resuming, for_replace)
            self._headerbar_stack.set_visible_child(self._search_header_bar)
        else:
            self._render_search_header_bar.enter(self.__render_view, resuming, type_to_search=False)
            self._headerbar_stack.set_visible_child(self._render_search_header_bar)

        if not self._top_bar_revealer.get_reveal_child():
            self.__show_header_bar()

    def __exit_search(self) -> None:
        """Stop searching within note."""
        self._headerbar_stack.set_visible_child(self._main_header_bar)
        self.__check_and_hide_bars_after_delay()
        self._sourceview.grab_focus()
        self._search_header_bar.exit()

    def __exit_render_search(self) -> None:
        self._headerbar_stack.set_visible_child(self._main_header_bar)
        self.__check_and_hide_bars_after_delay()
        self._render_search_header_bar.exit()

    def __toggle_render(self, force: bool = False) -> None:
        """Toggle whether showing rendered view."""

        if not self.__config_manager.markdown_render_enabled and not force:
            return

        if self._render_edit_stack.get_visible_child() == self._editor_scrolledwin:
            if self._headerbar_stack.get_visible_child() == self._search_header_bar:
                self.__exit_search()

            if self._bottom_bar_revealer.get_reveal_child():
                self._bottom_bar_revealer.set_visible(False)

            if self.__render_view is not None:
                self.__continue_loading_webkit()
            else:
                self._render_edit_stack.set_visible_child(self._render_loading)
                # Push the WebKit initialisation down the road a little, giving the loading view
                # time to initialise. A cleaner way to do this will be welcomed.
                GLib.timeout_add(50, self.__continue_loading_webkit)
        else:
            assert self.__render_view
            if self._headerbar_stack.get_visible_child() == self._render_search_header_bar:
                self.__exit_render_search()

            self._render_edit_button_stack.set_visible_child(self._render_button)
            self.__set_scroll_position(self.__render_view.scroll_position)
            self._sourceview.set_visible(True)
            self._render_edit_stack.set_visible_child(self._editor_scrolledwin)
            self.__check_and_terminate_webkit_process()
            self._bottom_bar_revealer.set_visible(True)
            self.__update_formatting_bar_visibility()
            self.__check_and_hide_bars_after_delay()
            self.focus_textview_if_editing()
            self._render_search_header_bar.disable_actions()
            self.__focus_mode_action.set_enabled(True)

    def __check_and_terminate_webkit_process(self) -> None:
        assert self.__render_view
        if (
            not self.__config_manager.markdown_render_enabled
            or not self.__config_manager.markdown_keep_webkit_process
        ):
            logging.debug("Terminating WebKit process")
            self.__render_view.terminate_web_process()
            # Reduce flash when next loading render view
            self.__render_view.set_visible(False)

    def __update_readonly_actions(self):
        editable = not not self.__note.read_only
        for action in ("rename", "edit-category", "clear-category", "delete-note"):
            self.__action_group.lookup_action(action).set_enabled(editable)

    def __update_scheme_and_dark_style(self) -> None:
        """Sync. dark style change to editor style."""
        buffer = self._sourceview.get_buffer()
        if not buffer:
            return

        style_manager = Adw.StyleManager.get_default()
        style_scheme_manager = GtkSource.StyleSchemeManager.get_default()

        scheme_id = self.__config_manager.editor_theme
        if style_manager.get_dark():
            scheme_id = f"{scheme_id}-dark"

        scheme = None
        if style_manager.get_high_contrast():
            high_contrast_id = f"{scheme_id}-high-contrast"
            scheme = style_scheme_manager.get_scheme(high_contrast_id)
            if not scheme:
                logging.debug(f"Scheme '{high_contrast_id}' doesn't exist")

        if scheme is None:
            scheme = style_scheme_manager.get_scheme(scheme_id)
        buffer.set_style_scheme(scheme)

    def __refresh_line_length_from_setting(self) -> None:
        length = self.__config_manager.line_length
        length_for_view = -1 if length == self.__config_manager.line_length_max else length
        self._sourceview.line_length = length_for_view

    def __push_font_updates(self) -> None:
        family = self.__get_editor_font_family()
        self._sourceview.update_font_family(family)

        if self.__html_generator is not None:
            family = self.__get_render_font_family()
            self.__html_generator.update_font_family(family)
            if self.__render_view:
                self.__render_view.update_style()

    def __get_render_font_family(self) -> str:
        return self.__get_font_family(self.__config_manager.markdown_use_monospace_font)

    def __get_editor_font_family(self) -> str:
        return self.__get_font_family(self.__config_manager.use_monospace_font)

    def __get_font_family(self, monospace: bool) -> str:
        style_manager = Adw.StyleManager.get_default()
        if monospace:
            font = style_manager.props.monospace_font_name
        else:
            font = style_manager.props.document_font_name
        font_description = Pango.font_description_from_string(font)
        return font_description.get_family()

    def __show_header_bar(self) -> None:
        """Show the headerbar."""
        self._top_bar_revealer.set_reveal_child(True)

    def __check_and_hide_bars_after_delay(self) -> None:
        if self.__hiding_bars_timeout is not None:
            return

        if self.__both_toolbars_hidden():
            return

        if self.__neither_toolbar_auto_hiding():
            return

        def callback():
            if self.active:
                self.__check_and_hide_header_bar()
                self.__check_and_hide_formatting_bar()
            self.__hiding_bars_timeout = None

        self.__hiding_bars_timeout = GLib.timeout_add(self.HEADERBAR_HIDE_DELAY, callback)

    def __both_toolbars_hidden(self) -> bool:
        return (
            not self._top_bar_revealer.get_reveal_child()
            and not self._bottom_bar_revealer.get_reveal_child()
        )

    def __neither_toolbar_auto_hiding(self) -> bool:
        return (
            self.__formatting_bar_visibility_setting
            == self.__header_bar_visibility_setting
            == HeaderBarVisibility.ALWAYS_VISIBLE
        )

    def __hide_header_bar(self) -> None:
        if self._headerbar_stack.get_visible_child() != self._main_header_bar:
            return
        elif self._menu_button.get_popover().get_visible():
            return

        if self._top_bar_revealer.get_reveal_child():
            self._top_bar_revealer.set_reveal_child(False)
            self.focus_textview_if_editing()

    def __show_formatting_bar(self) -> None:
        self._bottom_bar_revealer.set_reveal_child(True)

    def __hide_formatting_bar(self) -> None:
        self._bottom_bar_revealer.set_reveal_child(False)
        self.focus_textview_if_editing()

    def __check_and_hide_formatting_bar(self) -> bool:
        if not self._bottom_bar_revealer.get_reveal_child():
            return False

        if not self.__config_manager.markdown_detect_syntax:
            return False

        setting = self.__formatting_bar_visibility_setting
        hide = False
        if setting in (HeaderBarVisibility.AUTO_HIDE, HeaderBarVisibility.DISABLED):
            hide = True
        elif setting == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
            hide = self.get_root().is_fullscreen()

        if hide:
            self.__hide_formatting_bar()
        return hide

    def __check_and_show_formatting_bar(self) -> None:
        if self.__hiding_bars_timeout:
            GLib.source_remove(self.__hiding_bars_timeout)
            self.__hiding_bars_timeout = None
        if self._bottom_bar_revealer.get_reveal_child():
            return
        if not self.active or self._render_edit_stack.get_visible_child() == self.__render_view:
            return
        if not self.__config_manager.markdown_detect_syntax:
            return

        if self.__formatting_bar_visibility_setting != HeaderBarVisibility.DISABLED:
            self.__show_formatting_bar()

    def __check_and_hide_header_bar(self) -> None:
        if not self._top_bar_revealer.get_reveal_child():
            return

        hide = False
        if self.__header_bar_visibility_setting == HeaderBarVisibility.AUTO_HIDE:
            hide = True
        elif self.__header_bar_visibility_setting == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
            hide = self.get_root().is_fullscreen()

        if hide:
            self.__hide_header_bar()

    def __check_and_show_header_bar(self) -> None:
        if self.__hiding_bars_timeout:
            GLib.source_remove(self.__hiding_bars_timeout)
            self.__hiding_bars_timeout = None
        if not self._top_bar_revealer.get_reveal_child():
            self.__show_header_bar()

    def __update_for_formatting_bar_setting_change(self) -> None:
        self.__cache_formatting_bar_visibility()
        if self.active:
            self.__update_toolbar_underlay_margins()
            self.__update_formatting_bar_visibility()

    def __process_top_margin_input(self, x: float, y: float) -> None:
        location = Graphene.Point.init(Graphene.Point.alloc(), x, y)
        if not location.near(self.__pointer_position, 1.0):
            self.__pointer_position = location
            self.__check_and_show_header_bar()

    def __process_bottom_margin_input(self, x: float, y: float) -> None:
        self.__check_and_show_formatting_bar()
        location = Graphene.Point.init(Graphene.Point.alloc(), x, y)
        if not location.near(self.__pointer_position, 1.0):
            self.__pointer_position = location
            self.__check_and_show_formatting_bar()

    def __update_toolbar_underlay_margins(self) -> None:
        pad = self.__config_manager.editor_header_bar_visible_for_window_state
        if pad:
            margin = self._sourceview.padded_header_bar_height
        else:
            margin = 0
        if self._render_edit_stack.get_margin_top() != margin:
            self._render_edit_stack.set_margin_top(margin)

        pad = self.__config_manager.editor_formatting_bar_visible_for_window_state
        if pad:
            margin = self._formatting_header_bar.get_property("height-request")
        else:
            margin = 0
        if self._editor_scrolledwin.get_margin_bottom() != margin:
            self._editor_scrolledwin.set_margin_bottom(margin)

    def __update_formatting_bar_visibility(self) -> None:
        if not self.active or self._render_edit_stack.get_visible_child() == self.__render_view:
            return

        if not self.__config_manager.markdown_detect_syntax:
            return

        setting = self.__formatting_bar_visibility_setting
        if self._bottom_bar_revealer.get_reveal_child():
            self.__check_and_hide_formatting_bar()
        else:
            show = False
            if setting == HeaderBarVisibility.ALWAYS_VISIBLE:
                show = True
            elif setting == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
                show = not self.get_root().is_fullscreen()
            if show:
                self.__show_formatting_bar()

    def __update_for_header_bar_setting_change(self) -> None:
        self.__cache_header_bar_visibility()

        if not self.active:
            return

        setting = self.__header_bar_visibility_setting
        if self._top_bar_revealer.get_reveal_child():
            self.__check_and_hide_header_bar()
        else:
            show = False
            if setting == HeaderBarVisibility.ALWAYS_VISIBLE:
                show = True
            elif setting == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
                show = not self.get_root().is_fullscreen()
            if show:
                self.__show_header_bar()
        self.__update_toolbar_underlay_margins()

    def __cache_header_bar_visibility(self) -> None:
        self.__header_bar_visibility_setting = self.__config_manager.editor_header_bar_visibility

    def __cache_formatting_bar_visibility(self) -> None:
        self.__formatting_bar_visibility_setting = (
            self.__config_manager.editor_formatting_bar_visibility
        )

    def __continue_loading_webkit(self) -> None:
        assert self.__note

        self.ensure_webkit_initialised()

        assert self.__render_view

        if self.__sync_manager.authenticated:
            download_func = self.__sync_manager.download_attachment
        else:
            download_func = None
        self.__render_view.render(
            self.__note,
            export_format=None,
            attachment_download_func=download_func,
            scroll_position=self.__get_scroll_position(),
        )
        self._search_header_bar.disable_actions()
        self.__focus_mode_action.set_enabled(False)

    def __click_in_top_margin(self, gesture: Gtk.Gesture) -> bool:
        interpreted, _, y = gesture.get_point(None)
        if not interpreted:
            return False
        if not self._top_bar_revealer.get_reveal_child():
            return False

        return y < self._sourceview.get_property("top-margin")

    def __click_in_bottom_margin(self, gesture: Gtk.Gesture) -> bool:
        interpreted, _, y = gesture.get_point(None)
        if not interpreted:
            return False
        if not self._bottom_bar_revealer.get_reveal_child():
            return False

        return y > self._sourceview.get_height() - self._sourceview.get_property("top-margin")

    def __focus_editor(self) -> None:
        # Don't focus if showing render view
        if self._render_edit_stack.get_visible_child() != self._editor_scrolledwin:
            return

        self._sourceview.grab_focus()

    def __focus_header_bar(self) -> None:
        self.__show_header_bar()
        if self._headerbar_stack.get_visible_child() == self._rename_header_bar:
            self._rename_header_bar.focus()
        elif self._headerbar_stack.get_visible_child() == self._search_header_bar:
            self._search_header_bar.refocus_search_and_select()
        elif self._headerbar_stack.get_visible_child() == self._category_header_bar:
            self._category_header_bar.focus()
        elif self._headerbar_stack.get_visible_child() == self._render_search_header_bar:
            self._render_search_header_bar.focus()
        else:
            self._title_button.grab_focus()

    def __focus_formatting_bar(self) -> None:
        if not self.__config_manager.markdown_detect_syntax:
            return
        if self.__formatting_bar_visibility_setting == HeaderBarVisibility.DISABLED:
            return
        self.__show_formatting_bar()
        self._formatting_header_bar.focus()

    def __check_and_handle_link_click(self, x: float, y: float) -> None:
        link = None

        buffer_x, buffer_y = self._sourceview.window_to_buffer_coords(
            Gtk.TextWindowType.TEXT,
            x,
            y,
        )
        over_text, iter = self._sourceview.get_iter_at_location(buffer_x, buffer_y)
        if not over_text:
            logging.debug("Not over text for URL ctrl-click")
            return

        # Check and open basic link
        result = parse_any_url_at_iter(iter, http_only=True)
        if result:
            link = result.link
        else:
            # Check for a markdown inline and automatic links
            classes = iter.get_buffer().get_context_classes_at_iter(iter)
            if "inline-link" in classes:
                inline_result = parse_markdown_inline_link(iter)
                if inline_result:
                    link = inline_result.link
            elif "automatic-link" in classes:
                auto_result = parse_markdown_automatic_link(iter)
                if auto_result:
                    link = auto_result.link

        if link:
            logging.info(f"Opening URI from ctrl-click: {link}")
            webbrowser.open(link)
            self.__show_toast(_("Opening link in browser"))

    def __store_session_details(self) -> None:
        scroll_position = self._editor_scrolledwin.get_vadjustment().get_value()

        buffer = self._sourceview.get_buffer()
        insert_iter = buffer.get_iter_at_mark(buffer.get_insert())
        cursor_location = insert_iter.get_offset()

        if len(self.__previous_sessions) and self.__previous_sessions[-1].note == self.current_note:
            return

        details = EditorSessionDetails(
            self.current_note,
            scroll_position,
            cursor_location,
        )
        self.__previous_sessions.append(details)
        if len(self.__previous_sessions) > 2:
            self.__previous_sessions = self.__previous_sessions[-2:]

    def __flush_note_from_previous_sessions(self, note: Note) -> None:
        if self.__previous_sessions and self.__previous_sessions[-1].note == note:
            del self.__previous_sessions[-1]

    def __restore_session_location(self, session: EditorSessionDetails) -> None:
        self.__set_scroll_position(session.scroll_position)
        buffer = self._sourceview.get_buffer()
        cursor_iter = buffer.get_iter_at_offset(session.cursor_position)
        buffer.place_cursor(cursor_iter)

    def __show_outline(self) -> None:
        assert self.__note

        if self.__outline_dialog:
            self.__outline_dialog.close()

        self.__enable_actions(False)

        dialog = OutlineDialog()
        dialog.present(self)

        def populate(headings: list[OutlineHeading]) -> None:
            # This makes a crude assumption that if the locale is RTL the document contents are RTL.
            # Likely better than ignoring RTL but worse than knowing the document direction.
            rtl = self.get_default_direction() == Gtk.TextDirection.RTL
            dialog.populate(headings, rtl)

        self.__outline_generator.generate(self.__note, populate)

        def jump_to(dialog: OutlineDialog, line: int) -> None:
            if self._render_edit_stack.get_visible_child() == self._editor_scrolledwin:
                buffer = self._sourceview.get_buffer()
                success, line_iter = buffer.get_iter_at_line(line)
                buffer.place_cursor(line_iter)
                self._sourceview.jump_to_insertion_point()
            else:
                assert self.__render_view
                self.__render_view.scroll_to_heading(line)

        dialog.connect("jump-to", jump_to)

        def outline_closed(dialog: OutlineDialog) -> None:
            dialog.disconnect_by_func(jump_to)
            dialog.disconnect_by_func(outline_closed)
            self.__outline_dialog = None
            self.__enable_actions(True)

        dialog.connect("closed", outline_closed)
        self.__outline_dialog = dialog

    def __initialise_html_generator(self) -> None:
        if self.__html_generator:
            return

        from iotas.template_html_generator import TemplateHtmlGenerator

        self.__html_generator = TemplateHtmlGenerator(
            self.__config_manager, const.PKGDATADIR, self._sourceview.padded_header_bar_height
        )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gettext import gettext as _
from gi.repository import GObject

from enum import IntEnum, auto
from typing import Optional, Self

from iotas.note import Note


class CategorySpecialPurpose(IntEnum):
    ALL = auto()
    UNCATEGORISED = auto()


class Category(GObject.Object):
    # Translators: Title
    ALL_TITLE = _("All Notes")
    # Translators: Title
    UNCATEGORISED_TITLE = _("Uncategorized")

    def __init__(self, name: str, note_count: int) -> None:
        super().__init__()
        self.__parent = None
        self.name = name
        self.__sub_category_name = ""
        self.__child_note_count = 0
        self.__note_count_with_children = 0
        self.note_count = note_count
        self.__children: list[Category] = []
        self.__special_purpose: Optional[CategorySpecialPurpose] = None

    def includes_note(self, note: Note) -> bool:
        """Whether category or sub-category includes note.

        :param Note note: Note to check
        :returns: Whether included
        :rtype: bool
        """
        if note.locally_deleted:
            include = False
        elif self.__special_purpose == CategorySpecialPurpose.ALL:
            include = True
        elif self.__special_purpose == CategorySpecialPurpose.UNCATEGORISED:
            include = note.category == ""
        elif note.category == self.__name:
            include = True
        elif note.category.startswith(self.__name + "/"):
            include = True
        else:
            include = False
        return include

    def increase_note_count(self, amount: int = 1) -> None:
        """Increase note count.

        :param int amount: Amount to change by
        """
        self.note_count += amount
        if self.__parent is not None:
            self.__parent.increase_child_note_count(amount)

    def decrease_note_count(self, amount: int = 1) -> None:
        """Decrease note count.

        :param int amount: Amount to change by
        """
        self.note_count -= amount
        if self.__parent is not None:
            self.__parent.decrease_child_note_count(amount)

    def increase_child_note_count(self, amount: int = 1) -> None:
        """Increase children's note count.

        :param int amount: Amount to change by
        """
        self.child_note_count += amount
        if self.__parent is not None:
            self.__parent.increase_child_note_count(amount)

    def decrease_child_note_count(self, amount: int = 1) -> None:
        """Decrease children's note count.

        :param int amount: Amount to change by
        """
        self.child_note_count -= amount
        if self.__parent is not None:
            self.__parent.decrease_child_note_count(amount)

    def add_child(self, category: Category) -> None:
        """Link a child category.

        :param Category category: The new child
        """
        if category not in self.__children:
            self.__children.append(category)

    def remove_child(self, category: Category) -> None:
        """Unlink a child category.

        :param Category category: The child
        """
        if category in self.__children:
            self.__children.remove(category)

    def get_parent(self) -> Optional[Self]:
        """Get the parent category.

        :returns: The parent or None
        :rtype: Optional[Self]
        """
        return self.__parent

    def set_parent(self, category: Optional[Category]) -> None:
        """Set the parent category.

        :param Optional[Category] category: The parent or None
        """
        self.__parent = category
        if category is not None:
            self.__refresh_sub_category_name()

    # TODO remove with CategoryManager.recreate_category_for_expandable_change and Gtk v4.10+
    def duplicate(self) -> Self:
        """Duplicate the category.

        :returns: The duplicate
        :rtype: Self
        """
        new_category = Category(self.name, self.__note_count)
        new_category.set_parent(self.__parent)
        for child in self.__children:
            new_category.add_child(child)
            child.set_parent(new_category)
            new_category.increase_child_note_count(child.note_count)
        return new_category

    @GObject.Property(type=str, default="")
    def name(self) -> str:
        return self.__name

    @name.setter
    def set_name(self, value: str) -> None:
        self.__name = value
        self.__refresh_display_name()
        self.__sub_category_path = "/" in value
        if self.__sub_category_path:
            self.__refresh_sub_category_name()

    @GObject.Property(type=str, default="")
    def display_name(self) -> str:
        return self.__display_name

    @GObject.Property(type=int, default=0)
    def note_count(self) -> int:
        return self.__note_count

    @note_count.setter
    def set_note_count(self, value: int) -> None:
        self.__note_count = value
        self.note_count_with_children = self.child_note_count + value

    @GObject.Property(type=int, default=0)
    def child_note_count(self) -> int:
        return self.__child_note_count

    @child_note_count.setter
    def set_child_note_count(self, value: int) -> None:
        self.__child_note_count = value
        self.note_count_with_children = self.note_count + value

    @GObject.Property(type=int, default=0)
    def note_count_with_children(self) -> int:
        return self.__note_count_with_children

    @note_count_with_children.setter
    def set_note_count_with_children(self, value: int) -> None:
        self.__note_count_with_children = value

    @GObject.Property(type=int, default=False)
    def number_of_direct_children(self) -> int:
        return len(self.__children)

    @GObject.Property(type=int, default=None)
    def special_purpose(self) -> Optional[CategorySpecialPurpose]:
        return self.__special_purpose

    @special_purpose.setter
    def set_special_purpose(self, value: Optional[CategorySpecialPurpose]) -> None:
        self.__special_purpose = value

    @GObject.Property(type=str, default="")
    def sub_category_name(self) -> str:
        return self.__sub_category_name

    @GObject.Property(type=bool, default=False)
    def is_sub_category(self) -> bool:
        return self.__sub_category_path

    def __refresh_display_name(self):
        self.__display_name = Category.generate_display_name(self.__name)

    def __refresh_sub_category_name(self):
        if self.__parent is None and not self.__sub_category_path:
            return
        self.__sub_category_name = Category.generate_sub_category_name(self.__name)

    @staticmethod
    def generate_display_name(name: str) -> str:
        if "/" in name:
            return " / ".join(name.split("/"))
        else:
            return name

    @staticmethod
    def generate_sub_category_name(name: str) -> str:
        if "/" in name:
            return " / ".join(name.split("/")[1:])
        else:
            # Shouldn't occur
            return name



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod

from iotas.category import Category


class CategoryStorage(ABC):
    """Category storage interface."""

    @abstractmethod
    def get_all_categories(self) -> list[Category]:
        """Fetch all categories from the database.

        :return: List of categories
        :rtype: list[Category]
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gettext import ngettext
from gi.repository import Adw, Gdk, Gio, GObject, GLib, Gtk

import logging
import os
from tempfile import TemporaryDirectory
from typing import Callable, Optional

from iotas.attachment import Attachment
from iotas.config_manager import ConfigManager
import iotas.const as const
from iotas.export_utils import (
    build_batch_dirname,
    build_default_filename,
    get_container_export_dir,
    exporting_to_dir,
    parse_export_formats,
)
from iotas.note import Note
from iotas.webkit_pdf_exporter import WebKitPdfExporter


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/export_dialog.ui")
class ExportDialog(Adw.Dialog):
    __gtype_name__ = "ExportDialog"

    _main_stack: Gtk.Stack = Gtk.Template.Child()
    _headerbar_stack: Gtk.Stack = Gtk.Template.Child()
    _checking_attachments_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _format_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _downloading_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _exporting_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _success_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _failure_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _failure_headerbar_label: Gtk.Label = Gtk.Template.Child()
    _format_listbox: Gtk.ListBox = Gtk.Template.Child()
    _cancel_button: Gtk.Button = Gtk.Template.Child()
    _show_button: Gtk.Button = Gtk.Template.Child()
    _parsing: Adw.StatusPage = Gtk.Template.Child()
    _downloading: Adw.StatusPage = Gtk.Template.Child()
    _downloading_progress_bar: Gtk.ProgressBar = Gtk.Template.Child()
    _single_exporting: Gtk.Box = Gtk.Template.Child()
    _multi_exporting: Gtk.Box = Gtk.Template.Child()
    _multi_export_progress_bar: Gtk.ProgressBar = Gtk.Template.Child()
    _success: Gtk.Box = Gtk.Template.Child()
    _success_label: Gtk.Label = Gtk.Template.Child()
    _failure: Gtk.Box = Gtk.Template.Child()
    _failure_label: Gtk.Label = Gtk.Template.Child()
    _download_failure: Adw.StatusPage = Gtk.Template.Child()
    _retry_row: Adw.ButtonRow = Gtk.Template.Child()

    def __init__(
        self,
        notes: list[Note],
        html_generator,
        download_attachment: Optional[Callable[[Attachment], bool]],
        webkit_initialiser: Callable[[], None],
        note_content_populator: Optional[Callable[[Note], None]],
    ) -> None:
        super().__init__()

        self.__html_generator = html_generator
        self.__download_attachment = download_attachment
        self.__webkit_init = webkit_initialiser
        self.__note_content_populator = note_content_populator
        self.__config_manager = ConfigManager.get_default()

        # Any type for lazy loading
        self.__show_path: Optional[str] = None
        self.__custom_formats: dict[str, dict[str, str]] = {}

        self.__init_exporter()
        self.__exporter.set_notes(notes)

        controller = Gtk.EventControllerKey.new()
        controller.connect("key-pressed", self._on_key_pressed)
        self.add_controller(controller)

        GLib.idle_add(self.__setup)

        self.connect("closed", lambda _o: self.__on_close())

    def __setup(self) -> None:
        self._cancel_button.grab_focus()
        self.__load_extra_formats()

    @Gtk.Template.Callback()
    def _row_activated(self, _listbox: Gtk.ListBox, row: Adw.ButtonRow) -> None:
        format_id = row.get_title().lower()
        if format_id in self.__custom_formats:
            self.__exporter.file_extension = self.__custom_formats[format_id]["extension"]
            self.__exporter.out_format = self.__custom_formats[format_id]["format"]
        else:
            self.__exporter.file_extension = format_id
            self.__exporter.out_format = format_id
        self.__determine_output_location()

    @Gtk.Template.Callback()
    def _download_failure_row_activated(self, _listbox: Gtk.ListBox, row: Adw.ButtonRow) -> None:
        if row == self._retry_row:
            self.__check_and_transfer_missing_attachments()
        else:
            self.__exporter.start_export()

    @Gtk.Template.Callback()
    def _on_show_clicked(self, _button: Gtk.Button) -> None:
        launcher = Gtk.FileLauncher.new(Gio.File.new_for_uri(f"file://{self.__show_path}"))
        launcher.open_containing_folder()
        self.close()

    @Gtk.Template.Callback()
    def _on_cancel(self, _button: Gtk.Button) -> None:
        self.close()

    def _on_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if self._main_stack.get_visible_child() != self._single_exporting:
            if keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down):
                if self.get_property("focus-widget") == self._cancel_button:
                    row = self._format_listbox.get_row_at_index(0)
                    row.grab_focus()
                    return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def __on_note_exported(self, _obj: GObject.Object, progress: float) -> None:
        if self.__exporter.multi_export:
            self._multi_export_progress_bar.set_fraction(progress)

    def __on_batch_finished(self, _obj: GObject.Object, show_path: str) -> None:
        self._main_stack.set_visible_child(self._success)
        self._headerbar_stack.set_visible_child(self._success_headerbar)
        if self.__exporter.multi_export:
            # Translators: Description, {0} is a number greater than 1, {} is a format eg. PDF
            label = _("Exported {0} notes to {1}").format(
                len(self.__exporter.get_notes()), self.__exporter.out_format.upper()
            )
        else:
            # Translators: Description, {} is a format eg. PDF
            label = _("Exported to {}").format(self.__exporter.out_format.upper())
        self._success_label.set_label(label)
        self.__show_path = show_path

        # For some reason if grab focus is called directly from the idle it blocks clean exit
        def focus_show_button():
            self._show_button.grab_focus()

        GLib.idle_add(focus_show_button)

    def __on_export_failed(self, _obj: GObject.Object, reason: str) -> None:
        self.__report_failure(reason)

    def __on_save_dialog_finish(self, dialog: Gtk.FileDialog, task: Gio.Task) -> None:
        assert self.__exporter  # mypy
        try:
            file = dialog.save_finish(task)
        except GLib.GError as e:
            logging.warning(f"Couldn't export note: {e.message}")
            self.close()
        else:
            self.__exporter.location = file.get_path()
            logging.info(f"Got location {self.__exporter.location}")

            # Handle case where dir actually isn't writable, possibly just a sandbox dev
            # environment issue. Early 2026.
            notes = self.__exporter.get_notes()
            if self.__exporter.multi_export or exporting_to_dir(
                self.__exporter.out_format, notes[0]
            ):
                if not self.__ensure_dir_writable():
                    return

            if self.__exporter.location:
                self.__save_last_export_directory()

            self.__start_check_or_export()

    def __on_attachment_transfer_started(self, _obj: GObject.Object, count: int) -> None:
        self._downloading_progress_bar.set_fraction(0)

        # Translators: Description, notification, {} is a number
        msg = ngettext("Transferring {} attachment", "Transferring {} attachments", count).format(
            count
        )
        self._downloading.set_description(msg)
        self._headerbar_stack.set_visible_child(self._downloading_headerbar)
        self._main_stack.set_visible_child(self._downloading)
        # self.__failed_attachment_transfers = set()

    def __on_attachment_transferred(self, _obj: GObject.Object, progress: float) -> None:
        self._downloading_progress_bar.set_fraction(progress)

    def __on_finished_attachment_run(self, _obj: GObject.Object, count: int) -> None:
        if count:
            self.__show_attachment_download_failures(count)
        else:
            self.__exporter.start_export()

    def __on_export_started(self, _obj: GObject.Object) -> None:
        self._headerbar_stack.set_visible_child(self._exporting_headerbar)
        if self.__exporter.multi_export:
            self._main_stack.set_visible_child(self._multi_exporting)
        else:
            self._main_stack.set_visible_child(self._single_exporting)

    def __on_export_progress(self, _obj: GObject.Object, progress: float) -> None:
        if self.__exporter.multi_export:
            self._multi_export_progress_bar.set_fraction(progress)

    def __on_close(self) -> None:
        if self.__exporter:
            self.__exporter.disconnect_by_func(self.__on_attachment_transfer_started)
            self.__exporter.disconnect_by_func(self.__on_attachment_transferred)
            self.__exporter.disconnect_by_func(self.__on_finished_attachment_run)
            self.__exporter.disconnect_by_func(self.__on_export_started)
            self.__exporter.disconnect_by_func(self.__on_note_exported)
            self.__exporter.disconnect_by_func(self.__on_export_failed)
            self.__exporter.disconnect_by_func(self.__on_batch_finished)
            self.__exporter.cleanup()

    def __determine_output_location(self) -> None:
        filesystem_writable_path = self.__get_writable_non_container_dir()
        if filesystem_writable_path:
            dialog = Gtk.FileDialog.new()
            dialog.set_initial_folder(Gio.File.new_for_path(filesystem_writable_path))
            dialog.set_initial_name(self.__build_filename())
            # Translators: Button
            dialog.set_accept_label(_("Export"))
            window = self.get_parent().get_root()
            dialog.save(
                window,
                None,
                self.__on_save_dialog_finish,
            )
        else:
            logging.info(
                "Can't access user documents directory, exporting to location inside container"
            )
            path = self.__get_location_in_container()
            if path:
                self.__exporter.location = path
                self.__start_check_or_export()

    def __start_check_or_export(self) -> None:
        # If we're local only there is no attachment downloading, start the export
        if self.__download_attachment is None:
            self.__exporter.start_export()
        else:
            self.__check_and_transfer_missing_attachments()

    def __check_and_transfer_missing_attachments(self) -> None:
        self._headerbar_stack.set_visible_child(self._checking_attachments_headerbar)
        self._main_stack.set_visible_child(self._parsing)
        self.__exporter.check_and_transfer_missing_attachments()

    def __show_attachment_download_failures(self, count: int) -> None:
        # Translators: Description, notification, {} is a number
        msg = ngettext(
            "{} attachment failed to transfer", "{} attachments failed to transfer", count
        ).format(count)
        self._download_failure.set_description(msg)
        self._main_stack.set_visible_child(self._download_failure)
        self._headerbar_stack.set_visible_child(self._failure_headerbar)
        # Translators: Title
        self._failure_headerbar_label.set_label(_("Transfer Failed"))

    def __get_writable_non_container_dir(self) -> Optional[str]:
        last_dir = self.__config_manager.last_export_directory
        if last_dir != "" and self.__dir_exists_writable(last_dir):
            return last_dir
        elif self.__can_write_documents_dir():
            return self.__get_documents_dir()
        else:
            return None

    def __get_documents_dir(self) -> Optional[str]:
        return GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOCUMENTS)

    def __can_write_documents_dir(self) -> bool:
        documents_dir = self.__get_documents_dir()
        if documents_dir is None:
            return False
        else:
            return self.__dir_exists_writable(documents_dir)

    def __dir_exists_writable(self, path: str) -> bool:
        if not os.path.exists(path):
            return False

        try:
            with TemporaryDirectory(dir=path):
                return True
        except OSError:
            return False

    def __get_location_in_container(self) -> Optional[str]:
        export_dir = get_container_export_dir()

        if not os.path.exists(export_dir):
            logging.debug(f"Creating missing export directory {export_dir}")
            try:
                os.mkdir(export_dir)
            except OSError as e:
                logging.warning(f"Couldn't create exports directory within container: {e}")
                self.__report_failure(str(e))
                return None

        return os.path.join(export_dir, self.__build_filename())

    def __build_filename(self) -> str:
        if self.__exporter.multi_export:
            filename = build_batch_dirname()
        else:
            note = self.__exporter.get_notes()[0]
            if not note.content_loaded:
                # Always provided when exporting from list (otherwise non-null)
                assert self.__note_content_populator
                self.__note_content_populator(note)
            filename = build_default_filename(
                note, self.__exporter.out_format, self.__exporter.file_extension
            )
        return filename

    def __init_exporter(self) -> None:
        # More lazyloading to quicken startup
        from iotas.batch_exporter import BatchExporter
        from iotas.default_note_exporter import DefaultNoteExporter

        self.__pdf_exporter = WebKitPdfExporter(
            self.__webkit_init(), self.__config_manager.markdown_keep_webkit_process
        )

        single_note_exporter = DefaultNoteExporter(
            self.__pdf_exporter, self.__html_generator, const.PKGDATADIR
        )

        self.__exporter = BatchExporter(
            single_note_exporter, self.__download_attachment, self.__note_content_populator
        )
        self.__exporter.connect(
            "attachment-transfer-started", self.__on_attachment_transfer_started
        )
        self.__exporter.connect("attachment-transferred", self.__on_attachment_transferred)
        self.__exporter.connect(
            "finished-attachment-detection-transfer-run", self.__on_finished_attachment_run
        )
        self.__exporter.connect("export-started", self.__on_export_started)
        self.__exporter.connect("note-exported", self.__on_note_exported)
        self.__exporter.connect("export-failed", self.__on_export_failed)
        self.__exporter.connect("batch-finished", self.__on_batch_finished)

        self.__exporter.supporting_tex = self.__config_manager.markdown_tex_support

    def __load_extra_formats(self) -> None:
        raw_formats = self.__config_manager.extra_export_formats
        formats = parse_export_formats(raw_formats)
        if not formats:
            return

        for format in formats.values():
            button = Adw.ButtonRow()
            button.set_title(format["extension"].upper())
            button.set_end_icon_name("go-next-symbolic")
            self._format_listbox.append(button)

    def __save_last_export_directory(self) -> None:
        location = self.__exporter.location
        if os.path.isdir(location):
            dir_path = os.path.abspath(os.path.join(location, os.pardir))
        else:
            dir_path = os.path.dirname(location)
        self.__config_manager.last_export_directory = dir_path

    def __ensure_dir_writable(self) -> bool:
        success = True
        if self.__exporter.multi_export:
            path = self.__exporter.location
        else:
            path = os.path.dirname(self.__exporter.location)
        if not self.__dir_exists_writable(path):
            logging.info("Can't write to selected path, changing path to within data dir/container")
            path = self.__get_location_in_container()
            if path:
                self.__exporter.location = path
            else:
                success = False
        return success

    def __report_failure(self, reason: str) -> None:
        self._main_stack.set_visible_child(self._failure)
        self._headerbar_stack.set_visible_child(self._failure_headerbar)
        # Translators: Description, {} is a format eg. PDF
        message = _("Failed to export to {}").format(self.__exporter.out_format.upper())
        if reason != "":
            message = message + "\n\n" + reason
        logging.warning(message)
        self._failure_label.set_label(message)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import itertools
from typing import Optional

from iotas.backup_storage import BackupStorage
from iotas.category import Category, CategorySpecialPurpose
from iotas.category_storage import CategoryStorage
from iotas.database import Database, DbCursor
from iotas.note import Note, DirtyFields


class NoteDatabase(BackupStorage, CategoryStorage):

    def __init__(self, db_base: Database) -> None:
        self.__db_base = db_base

    def add_note(self, note: Note) -> None:
        """Add a new note to the database.

        :param Note note: Note to add
        """
        with DbCursor(self.__db_base, True) as sql:
            result = sql.execute(
                "INSERT INTO note (title, content, last_modified, remote_id,"
                "                  dirty, etag, locally_deleted, favourite,"
                "                  category, excerpt, read_only)"
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    note.title,
                    note.content,
                    note.last_modified,
                    note.remote_id,
                    note.dirty,
                    note.etag,
                    False,
                    note.favourite,
                    note.category,
                    note.excerpt,
                    note.read_only,
                ),
            )
            note.id = result.lastrowid

    def populate_note_content(self, note: Note) -> None:
        """Load note content for provided note.

        :param Note note: Note to load for
        """
        if note.content_loaded:
            return
        if note.id < 0:
            return
        with DbCursor(self.__db_base) as sql:
            request = "SELECT content FROM note WHERE rowid=?"
            result = sql.execute(request, (note.id,))
            row = result.fetchone()
            if row is not None:
                note.content = row[0]

    def create_duplicate_note(self, note: Note, reason: str) -> Note:
        """Create a duplicate note with a prefixed title.

        :param Note note: The note to duplicate
        :param str reason: The reason for the duplication, which is prefixed on the title
        """
        new_note = note.duplicate()
        new_note.title = f"{reason} - {note.title}"
        new_note.flag_changed()
        self.add_note(new_note)
        return new_note

    def delete_all_clean_synced_notes(self) -> None:
        """Delete all notes which are on the server and don't have local updates."""
        with DbCursor(self.__db_base, True) as sql:
            sql.execute("DELETE FROM note WHERE remote_id <> -1 AND DIRTY=0")

    def delete_note(self, note_id: int) -> None:
        """Delete a note.

        :param int note_id: The note identifier (row id, not remote id)
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute("DELETE FROM note WHERE rowid=?", (note_id,))

    def get_all_notes(self, load_content: bool = False) -> list[Note]:
        """Fetch all notes from the database.

        :param bool load_content: Whether to load the content for each note
        :return: List of notes
        :rtype: list[Note]
        """
        with DbCursor(self.__db_base) as sql:
            request = (
                "SELECT title, excerpt, last_modified, favourite, category, remote_id,"
                "       dirty, etag, rowid, locally_deleted, read_only"
            )
            if load_content:
                request += ", content"
            request += " FROM note"
            result = sql.execute(request)
            return [NoteDatabase.db_row_to_note(row) for row in result]

    def get_all_categories(self) -> list[Category]:
        """Fetch all categories from the database.

        :return: List of categories
        :rtype: list[Category]
        """
        with DbCursor(self.__db_base) as sql:
            request = (
                "SELECT category, count() AS count "
                "FROM note "
                "WHERE locally_deleted=0 "
                "GROUP BY category "
                "ORDER BY LOWER(category)"
            )
            result = sql.execute(request)
            categories = [NoteDatabase.db_row_to_category(row) for row in result]
            total = 0
            uncategorised_found = False
            for category in categories:
                total += category.note_count
                if category.name == Category.UNCATEGORISED_TITLE:
                    uncategorised_found = True
            all_category = Category(Category.ALL_TITLE, total)
            all_category.special_purpose = CategorySpecialPurpose.ALL
            categories.insert(0, all_category)

            # Create an empty uncategorised entry if we don't have any uncategorised notes
            # (most commonly happening for new users)
            if not uncategorised_found:
                uncategorised_category = Category(Category.UNCATEGORISED_TITLE, 0)
                uncategorised_category.special_purpose = CategorySpecialPurpose.UNCATEGORISED
                categories.insert(1, uncategorised_category)

            return categories

    def get_all_notes_count(self, include_locally_deleted: bool = True) -> int:
        """Fetch the number of notes in the database.

        :param bool include_locally_deleted: Whether to include locally deleted notes in the count
        :return: The count
        :rtype: int
        """
        with DbCursor(self.__db_base) as sql:
            request = "SELECT count() FROM note"
            if not include_locally_deleted:
                request += " locally_deleted=0"
            result = sql.execute(request)
            row = result.fetchone()
            return row[0]

    def get_notes_by_ids(self, db_ids: list[int]) -> list[Note]:
        """Fetch notes from the database by db ids.

        :param list[int] db_ids: Database id of the note
        :return: The notes, in the same order as the provided ids
        :rtype: list[Note]
        """
        if len(db_ids) == 0:
            return []
        with DbCursor(self.__db_base) as sql:
            request = (
                "SELECT title, excerpt, last_modified, favourite, category, remote_id,"
                "       dirty, etag, rowid, locally_deleted, content "
                f"FROM note WHERE id IN ({','.join(['?']*len(db_ids))})"
            )
            result = sql.execute(request, db_ids)
            as_dict = {row[8]: NoteDatabase.db_row_to_note(row) for row in result}
            return [as_dict[note_id] for note_id in db_ids]

    def get_all_notes_remote_ids(self) -> list[int]:
        """Fetch remote ids for all remotely synced notes.

        :return: List of ids
        :rtype: list[int]
        """
        with DbCursor(self.__db_base) as sql:
            request = "SELECT remote_id FROM note WHERE remote_id <> -1"
            result = sql.execute(request)
            return list(itertools.chain(*result))

    def get_note_by_remote_id(self, remote_id: int) -> Optional[Note]:
        """Fetch note from the database by remote id.

        :param int remote_id: Remote id of the note
        :return: The note, or None
        :rtype: Optional[Note]
        """
        with DbCursor(self.__db_base) as sql:
            request = (
                "SELECT title, excerpt, last_modified, favourite, category, remote_id,"
                "       dirty, etag, rowid, locally_deleted, content "
                "FROM note WHERE remote_id=? "
            )
            result = sql.execute(request, (remote_id,))
            row = result.fetchone()
            if row is not None:
                i = NoteDatabase.db_row_to_note(row)
                return i
            return None

    def get_note_by_title(self, title: str) -> Optional[Note]:
        """Fetch note from the database by title.

        :param str title: Search title
        :return: The note, or None
        :rtype: Optional[Note]
        """
        with DbCursor(self.__db_base) as sql:
            request = (
                "SELECT title, excerpt, last_modified, favourite, category, remote_id,"
                "       dirty, etag, rowid, locally_deleted, content "
                "FROM note WHERE title=? "
            )
            result = sql.execute(request, (title,))
            row = result.fetchone()
            if row is not None:
                i = NoteDatabase.db_row_to_note(row)
                return i
            return None

    def get_note_failed_pushes(self, note: Note) -> Optional[str]:
        """Fetch note failed pushes from the database.

        :param Note note: The note
        :return: Failed pushes as JSON string, or None
        :rtype: str, optional
        """
        with DbCursor(self.__db_base) as sql:
            request = "SELECT failed_pushes FROM note WHERE id = ?"
            result = sql.execute(request, (note.id,))
            row = result.fetchone()
            if row is None:
                return None
            else:
                return row[0]

    def note_needs_update(self, remote_id: int, etag: str) -> bool:
        """Whether a note needs updating from a remote update.

        :param int remote_id: The remote id for the note
        :param str etag: The etag in the update
        :return: Whether an update is required
        :rtype: bool
        """
        with DbCursor(self.__db_base) as sql:
            request = "SELECT 1 FROM note WHERE remote_id=? AND etag<>?"
            result = sql.execute(
                request,
                (
                    remote_id,
                    etag,
                ),
            )
            return result.fetchone() is not None

    def persist_note_category(self, note: Note) -> None:
        """Persist category field for note to database.

        :param Note note: Note to update
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note SET dirty=?, category=? WHERE rowid=?",
                (note.dirty, note.category, note.id),
            )

    def persist_note_dirty(self, note: Note) -> None:
        """Persist dirty field for note to database.

        :param Note note: Note to update
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note SET dirty=? WHERE rowid=?",
                (note.dirty, note.id),
            )

    def persist_note_editable(self, note: Note) -> None:
        """Save editable fields for provided note to the database.

        Updates title, content, category, excerpt, last_modified, favourite and dirty

        :param Note note: Note to update
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note "
                "SET title=?, content=?, excerpt=?,"
                "    last_modified=?, favourite=?, dirty=?, "
                "    category=? "
                "WHERE rowid=?",
                (
                    note.title,
                    note.content,
                    note.excerpt,
                    note.last_modified,
                    note.favourite,
                    note.dirty,
                    note.category,
                    note.id,
                ),
            )

    def persist_note_favourite(self, note: Note) -> None:
        """Persist favourite field for note to database.

        :param Note note: Note to update
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note SET dirty=?, favourite=? WHERE rowid=?",
                (note.dirty, note.favourite, note.id),
            )

    def persist_note_etag(self, note: Note) -> None:
        """Persist etag field for note to database.

        :param Note note: Note to update
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note SET etag=? WHERE rowid=?",
                (note.etag, note.id),
            )

    def persist_note_locally_deleted(self, note: Note, value: bool) -> None:
        """Persist whether a note is locally deleted.

        :param Note note: The note to update
        :param bool value: The value to set
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note SET locally_deleted=?, dirty=1 WHERE rowid=?",
                (value, note.id),
            )

    def persist_note_remote_meta_and_sanitisable_fields(
        self, note: Note, clear_dirty: bool
    ) -> None:
        """Persist remote metadata and server sanitisable fields to database for provided note.

        :param Note note: Note to update
        :param bool clear_dirty: Whether to clear the dirty flag
        """
        with DbCursor(self.__db_base, True) as sql:
            request = "UPDATE note SET last_modified=?, remote_id=?, etag=?, title=?, category=?"
            request += ", failed_pushes=NULL"
            if clear_dirty:
                request += ", dirty=0"
            request += " WHERE rowid=?"
            sql.execute(
                request,
                (
                    note.last_modified,
                    note.remote_id,
                    note.etag,
                    note.title,
                    note.category,
                    note.id,
                ),
            )

    def persist_note_selective(self, note: Note, updated_fields: DirtyFields) -> None:
        """Persist local note based on remote changes.

        :param Note note: The note to update
        :param DirtyFields updated_fields: Which fields to update
        """
        if updated_fields.empty():
            return
        with DbCursor(self.__db_base, True) as sql:
            request = "UPDATE note SET"
            fields = []
            if updated_fields.title:
                request += " title=?"
                fields.append(note.title)
            if updated_fields.content:
                if len(fields) > 0:
                    request += ","
                request += " content=?"
                fields.append(note.content)
            if updated_fields.excerpt:
                if len(fields) > 0:
                    request += ","
                request += " excerpt=?"
                fields.append(note.excerpt)
            if updated_fields.favourite:
                if len(fields) > 0:
                    request += ","
                request += " favourite=?"
                fields.append(note.favourite)
            if updated_fields.category:
                if len(fields) > 0:
                    request += ","
                request += " category=?"
                fields.append(note.category)
            if updated_fields.etag:
                if len(fields) > 0:
                    request += ","
                request += " etag=?"
                fields.append(note.etag)
            if updated_fields.last_modified:
                if len(fields) > 0:
                    request += ","
                request += " last_modified=?"
                fields.append(note.last_modified)
            if updated_fields.read_only:
                if len(fields) > 0:
                    request += ","
                request += " read_only=?"
                fields.append(note.read_only)
            if updated_fields.remote_id:
                if len(fields) > 0:
                    request += ","
                request += " remote_id=?"
                fields.append(note.remote_id)
            if updated_fields.locally_deleted:
                if len(fields) > 0:
                    request += ","
                request += " locally_deleted=?"
                fields.append(note.locally_deleted)
            request += " WHERE id=?"
            fields.append(note.id)
            sql.execute(request, fields)

    def persist_note_failed_pushes(self, note: Note, failed_pushes: Optional[str]) -> None:
        """Persist note failed pushes to database.

        :param Note note: Note to update
        :param Optional[str] failed_pushes: Failed pushes, as JSON string
        """
        with DbCursor(self.__db_base, True) as sql:
            sql.execute(
                "UPDATE note SET failed_pushes=? WHERE rowid=?",
                (failed_pushes, note.id),
            )

    def search_notes(self, term: str, restrict_to: list[int] = [], sort: bool = False) -> list[int]:
        """Search the notes.

        :param str term: The search term
        :param list[int] restrict_to: Database ids for notes to restrict to
        :return: A list of ids for any matching notes
        :rtype: list[int]
        """
        with DbCursor(self.__db_base) as sql:
            filters = []
            search = term.replace('"', '""')
            search = search.replace("'", "''")
            request = "SELECT i.rowid FROM note_fts_idx i "
            if sort:
                request += " INNER JOIN note n ON i.rowid = n.id AND n.locally_deleted = 0 "
            request += " WHERE  "
            if len(restrict_to) > 0:
                request += f" id IN ({','.join(['?']*len(restrict_to))}) AND"
                filters.extend(restrict_to)
            request += f" note_fts_idx MATCH '\"{search}\"*' "
            if sort:
                request += " ORDER BY last_modified DESC"

            result = sql.execute(request, filters)
            return list(itertools.chain(*result))

    @staticmethod
    def db_row_to_note(row: tuple) -> Note:
        """Create Note from database row.

        :param tuple row: The database row
        :return: The note
        :rtype: Note
        """
        i = Note()
        i.title = row[0]
        i.excerpt = row[1]
        i.last_modified = row[2]
        i.favourite = row[3]
        i.category = row[4]
        i.remote_id = row[5]
        i.dirty = row[6]
        i.etag = row[7]
        i.id = row[8]
        i.locally_deleted = row[9]
        i.read_only = row[10]
        if len(row) > 11:
            i.content = row[11]
        return i

    @staticmethod
    def db_row_to_category(row: tuple) -> Category:
        """Create Category from database row.

        :param tuple row: The database row
        :return: The category
        :rtype: Category
        """
        i = Category(row[0], row[1])
        if i.name == "":
            i.name = Category.UNCATEGORISED_TITLE
            i.special_purpose = CategorySpecialPurpose.UNCATEGORISED
        return i



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib, GObject

from collections import OrderedDict
from threading import Thread
from typing import Optional

from iotas.category_storage import CategoryStorage
from iotas.category import Category, CategorySpecialPurpose
from iotas.category_list_model import CategoryListModel
from iotas.category_treeview_list_store import CategoryTreeViewListStore


class CategoryManager(GObject.Object):
    __gtype_name__ = "CategoryManager"
    __gsignals__ = {
        "initial-load-complete": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, db: CategoryStorage) -> None:
        super().__init__()
        self.__db = db
        self.__list_model = CategoryListModel()
        self.__tree_store = CategoryTreeViewListStore()
        self.__all_category: Optional[Category] = None
        self.__uncategorised_category: Optional[Category] = None
        self.__categories: OrderedDict[str, Category] = OrderedDict()
        self.__have_sub_category = False

    def populate(self) -> None:
        """Populate categories from database."""

        def chain_population_to_stores() -> None:
            self.__list_model.populate(self.__get_with_special())
            # Lean on the sorting in the list store to populate the tree store with its sorted
            # list including parent categories
            categories = []
            for i in range(2, self.__list_model.get_n_items()):
                categories.append(self.__list_model.get_item(i))
            self.__tree_store.populate(categories)
            self.emit("initial-load-complete")

        def thread_do() -> None:
            categories = self.__db.get_all_categories()
            for category in categories:
                if category.special_purpose == CategorySpecialPurpose.ALL:
                    self.__all_category = category
                elif category.special_purpose == CategorySpecialPurpose.UNCATEGORISED:
                    self.__uncategorised_category = category
                else:
                    self.__categories[category.name] = category
                    if category.is_sub_category:
                        self.__ensure_parents(category, False)
                        self.have_sub_category = True

            GLib.idle_add(chain_population_to_stores)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def note_added(self, new_category: str) -> None:
        """Make updates for a created note.

        :param Optional[str] new_category: The category a note has been added to
        """
        self.note_category_changed(None, new_category)

    def note_category_changed(
        self, old_category: Optional[str], new_category: Optional[str]
    ) -> None:
        """Make updates for a note changing category.

        :param Optional[str] old_category: The category a note has been removed from
        :param Optional[str] new_category: The category a note has been added to
        """
        if old_category is not None:
            self.__decrement_category(old_category, new_category is None)

        if new_category is not None:
            if new_category != "":
                category_updated = self.__increment_category(new_category)
                if not category_updated:
                    self.__add_new_category(new_category)
            else:
                # Moving into uncategorised
                self.__increment_uncategorised()
            if old_category is None:
                # Handle new note creation
                self.__increment_all()

        if old_category is not None:
            self.__handle_possible_emptied_category(old_category)

    def note_deleted(self, old_category: str) -> None:
        """Make updates for a deleted note.

        :param Optional[str] old_category: The category a note has been removed from
        """
        self.note_category_changed(old_category, None)

    # TODO with Gtk v4.10+ replace this with a binding from the model count to TreeExpander's
    # hide-expander
    def recreate_category_for_expandable_change(self, category: Category) -> None:
        """Recreate the provided category.

        Triggers an update in the model when a category changes in and out of leaf state.

        :param Category category: The category to recreate
        """
        parent = category.get_parent()
        if parent is not None:
            parent.remove_child(category)
        new_category = category.duplicate()

        del self.__categories[category.name]
        self.__list_model.remove(category)
        self.__tree_store.delete(category)

        self.__categories[category.name] = new_category
        self.__list_model.add(new_category)
        self.__tree_store.add(new_category)
        if parent is not None:
            parent.add_child(new_category)

    def get_normal_category(self, name: str) -> tuple[Category | None, int]:
        return self.__list_model.get_non_special(name)

    def get_special_purpose_category(self, purpose: CategorySpecialPurpose) -> tuple[Category, int]:
        return self.__list_model.get_special(purpose)

    @GObject.Property(type=bool, default=False)
    def have_sub_category(self) -> bool:
        return self.__have_sub_category

    @have_sub_category.setter
    def set_have_sub_category(self, value: bool) -> None:
        self.__have_sub_category = value

    @GObject.Property(type=CategoryListModel, default=False)
    def list_model(self) -> CategoryListModel:
        return self.__list_model

    @GObject.Property(type=CategoryTreeViewListStore, default=False)
    def tree_store(self) -> CategoryTreeViewListStore:
        return self.__tree_store

    @GObject.Property(type=Category, default=None)
    def all_category(self) -> Optional[Category]:
        return self.__all_category

    def __get_with_special(self) -> list[Category]:
        assert self.__uncategorised_category
        assert self.__all_category
        out = list(self.__categories.values())
        out.insert(0, self.__uncategorised_category)
        out.insert(0, self.__all_category)
        return out

    def __add_new_category(self, category_name: str) -> None:
        category = Category(category_name, 1)
        self.__categories[category_name] = category
        parent = None
        if category.is_sub_category:
            self.__ensure_parents(category, True)
            parent = category.get_parent()
            self.have_sub_category = True
        self.__list_model.add(category)
        self.__tree_store.add(category)
        # If parent has just changed to expandable, recreate it
        if parent is not None and parent.number_of_direct_children == 1:
            self.recreate_category_for_expandable_change(parent)

    def __increment_category(self, category_name: str) -> bool:
        category_updated = False
        if category_name in self.__categories:
            self.__categories[category_name].increase_note_count()
            category_updated = True
        return category_updated

    def __decrement_category(self, old_category_name: str, note_removed: bool) -> None:
        assert self.__uncategorised_category
        assert self.__all_category
        if old_category_name == "":
            # Moving out of uncategorised
            self.__uncategorised_category.decrease_note_count()
        else:
            if old_category_name in self.__categories:
                self.__categories[old_category_name].decrease_note_count()

        # Handle deletion
        if note_removed:
            self.__all_category.decrease_note_count()

    def __handle_possible_emptied_category(self, old_category: str) -> None:
        if old_category not in self.__categories:
            return

        category = self.__categories[old_category]
        if category.note_count_with_children == 0:
            delete_stack = [category]
            parent = category.get_parent()
            while parent is not None and parent.note_count_with_children == 0:
                delete_stack.append(parent)
                parent = parent.get_parent()
            delete_stack.reverse()

            delete_parent = delete_stack[0].get_parent()
            if delete_parent is not None:
                # If we reach a node to be removed that has a parent with other notes
                # (on itself or siblings) remove this node from its parent
                delete_parent.remove_child(delete_stack[0])

            # Iterate deleting
            for parent in delete_stack:
                del self.__categories[parent.name]
                self.__list_model.remove(parent)
                self.__tree_store.delete(parent)

        self.__refresh_have_sub_category()

    def __increment_uncategorised(self) -> None:
        assert self.__uncategorised_category
        self.__uncategorised_category.increase_note_count()

    def __increment_all(self) -> None:
        assert self.__all_category
        self.__all_category.increase_note_count()

    def __ensure_parents(self, category: Category, populate_models: bool) -> None:
        parents = category.name.split("/")[:-1]
        full_path_parents = []
        parts = []
        for parent_name in parents:
            parts.append(parent_name)
            full_path_parents.append("/".join(parts))

        level_child = category
        for parent_name in reversed(full_path_parents):
            if level_child.get_parent() is None:
                # Check for existing
                if parent_name in self.__categories:
                    parent = self.__categories[parent_name]
                else:
                    # Create if one doesn't exist
                    parent = Category(parent_name, 0)
                    self.__categories[parent_name] = parent
                    if populate_models:
                        self.__list_model.add(parent)
                        self.__tree_store.add(parent)
                level_child.set_parent(parent)

                # Add child
                parent.add_child(level_child)

                # Update child note count
                parent.increase_child_note_count(level_child.note_count_with_children)

            # Prepare as child for next level
            level_child = parent

    def __refresh_have_sub_category(self):
        have = False
        for category in self.__categories.values():
            if category.is_sub_category:
                have = True
                break
        if self.have_sub_category != have:
            self.have_sub_category = have



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, GLib

from datetime import datetime, timedelta
import json
import logging
from threading import Thread
import time
from typing import Optional

from iotas.attachment_helpers import delete_attachments_for_note
from iotas.category import Category
from iotas.category_manager import CategoryManager
from iotas.config_manager import ConfigManager
from iotas.note import Note, DirtyFields
from iotas.note_database import NoteDatabase
from iotas.note_list_model import (
    NoteListModelBase,
    NoteListModelCategoryFiltered,
    NoteListModelFavourites,
    NoteListModelTimeFiltered,
    NoteListModelSearch,
)
from iotas.sync_result import ContentPushSyncResult, FailedPush


class NoteManager(GObject.Object):
    __gtype_name__ = "NoteManager"
    __gsignals__ = {
        "initial-load-complete": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "new-note-persisted": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "sync-requested": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    LOAD_ALL_NOTES_TOTAL_THRESHOLD = 50
    LOAD_ALL_NOTES_LAST_TWO_MONTHS_THRESHOLD = 6
    EDITING_NOTE_SAVE_INTERVAL = 1000

    def __init__(
        self,
        db: NoteDatabase,
        category_manager: CategoryManager,
    ) -> None:
        super().__init__()
        self.__db = db
        self.__category_manager = category_manager

        self.__base_model = NoteListModelBase()

        self.__category_filtered_model = NoteListModelCategoryFiltered(self.__base_model)
        self.__favourites_model = NoteListModelFavourites(self.__category_filtered_model)
        self.__today_model = NoteListModelTimeFiltered(self.__category_filtered_model)
        self.__yesterday_model = NoteListModelTimeFiltered(self.__category_filtered_model)
        self.__week_model = NoteListModelTimeFiltered(self.__category_filtered_model)
        self.__month_model = NoteListModelTimeFiltered(self.__category_filtered_model)
        self.__last_month_model = NoteListModelTimeFiltered(self.__category_filtered_model)
        self.__older_notes_model: Optional[NoteListModelTimeFiltered] = None
        self.__search_model: Optional[NoteListModelSearch] = None

        self.__save_timeout_id = None
        self.__deleted_notes: list[Note] = []

    def initiate_model_from_db(self) -> None:
        """Populate the models from the database."""

        def add_to_model(notes: list[Note]) -> None:
            start_time = time.time()
            self.add_notes_to_model(notes)
            logging.debug("Add to model took %.2fs", time.time() - start_time)
            self.emit("initial-load-complete")

        def thread_do() -> None:
            start_time = time.time()
            notes = self.__db.get_all_notes()
            GLib.idle_add(add_to_model, notes)
            logging.debug("Get from DB took %.2fs", time.time() - start_time)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def initiate_older_notes_model(self) -> NoteListModelTimeFiltered:
        """Initiate the older notes model."""
        self.__older_notes_model = NoteListModelTimeFiltered(self.__category_filtered_model)
        return self.__older_notes_model

    def initiate_search_model(self) -> NoteListModelSearch:
        """Initiate the search model."""
        self.__search_model = NoteListModelSearch(self.__base_model)
        return self.__search_model

    def add_notes_to_model(self, notes: list[Note]) -> None:
        """Add new notes to model.

        :param list[Note] notes: The notes to add
        """
        self.__base_model.add_notes(notes)

    def remove_notes_from_model(self, notes: list[Note]) -> None:
        """Remove notes from the model.

        :param list[Note] notes: The notes to remove
        """
        self.__base_model.remove_notes(notes)

    def create_note(self, current_category: Category) -> Note:
        """Create a new note.

        :param Category category: The category currently active in the index
        """
        note = Note(True)
        note.content = ""
        note.flag_changed()
        if current_category.special_purpose is None:
            note.category = current_category.name
        config_manager = ConfigManager.get_default()
        if config_manager.first_start:
            config_manager.first_start = False
        self.__category_manager.note_added(note.category)
        return note

    def delete_notes(self, notes: list[Note], provide_undo: bool) -> None:
        """Delete notes from the database and model.

        :param list[Note] notes: The notes
        :param bool provide_undo: Whether to provide ability to undo
        """
        if provide_undo:
            for note in self.__deleted_notes:
                delete_attachments_for_note(note)
            self.__deleted_notes = []

        for note in notes:
            if provide_undo and not note.content_loaded:
                self.__db.populate_note_content(note)

            note.locally_deleted = True
            if note.remote_id >= 0 and provide_undo:
                self.__db.persist_note_locally_deleted(note, True)
            else:
                self.__db.delete_note(note.id)
                self.remove_notes_from_model([note])

            if provide_undo:
                self.__deleted_notes.append(note)
            self.__category_manager.note_deleted(note.category)

            # Delay attachment removal for local deletions, allowing for the undo period
            if not provide_undo:
                delete_attachments_for_note(note)

    def fetch_note_by_remote_id(self, remote_id: int) -> Optional[Note]:
        """Fetch a note by remote id"

        :param int remote_id: The remote id of the note
        """
        return self.__base_model.fetch_note_by_remote_id(remote_id)

    def fetch_note_by_id(self, db_id: int) -> Optional[Note]:
        """Fetch a note by database id"

        :param int db_id: The database id of the note
        """
        return self.__base_model.fetch_note_by_db_id(db_id)

    def invalidate_sort(self) -> None:
        """Invalidate the sorter."""
        self.__base_model.invalidate_sort()

    def update_note_post_sync(self, note: Note, sync_result: ContentPushSyncResult) -> None:
        """Update note after remote sync.

        :param Note note: The note
        :param ContentPushSyncResult sync_result: The sync result
        """
        note.dirty = False
        note.repopulate_meta_from_sync_update(sync_result.data)
        clear_dirty = not note.dirty_while_saving
        self.__db.persist_note_remote_meta_and_sanitisable_fields(note, clear_dirty)
        self.__base_model.ensure_remote_map_entry(note)

    def on_deletion_undo_elapsed(self) -> None:
        """Push local deletions to server and flush local queue upon undo lapse."""
        had_remote = False
        for note in self.__deleted_notes:
            delete_attachments_for_note(note)
            if note.has_remote_id:
                had_remote = True
        self.__deleted_notes = []
        if had_remote:
            self.emit("sync-requested")

    def undo_deletion(self) -> bool:
        """Undo a pending deletion.

        :return: Whether any notes were restored
        :rtype: bool
        """
        if not self.__deleted_notes:
            return False

        for note in self.__deleted_notes:
            note.locally_deleted = False
            if note.remote_id >= 0:
                self.__db.persist_note_locally_deleted(note, False)
            else:
                note.id = -1
                note.last_modified = int(time.time())
                self.__db.add_note(note)
                self.add_notes_to_model([note])
            self.__category_manager.note_added(note.category)
        self.__deleted_notes = []
        return True

    def apply_deleted_notes_attachment_deletion(self) -> None:
        """Delete attachments for any locally deleted notes awaiting undo timeout.

        Ensures the removal of attachments for deleted notes if exiting before the undo toast has
        lapsed.
        """
        for note in self.__deleted_notes:
            delete_attachments_for_note(note)
        self.__deleted_notes = []

    def persist_note_while_editing(self, note: Note, update_excerpt: bool) -> None:
        """Persist note to database on separate thread while editing.

        :param Note note: The note
        :param bool update_excerpt: Whether to update excerpt
        """
        if self.__save_timeout_id is None:
            if note.new_and_empty:
                return
            self.__save_timeout_id = GLib.timeout_add(
                self.EDITING_NOTE_SAVE_INTERVAL,
                self.__persist_note_on_thread,
                note,
                update_excerpt,
            )

    def persist_note_sync(
        self, note: Note, allow_empty: bool = False, force_update_excerpt: bool = False
    ) -> None:
        """Persist note to database on main thread.

        :param Note note: The note
        :param bool allow_empty: Whether to store even when new and empty
        :param bool force_update_excerpt: Whether to update the excerpt, even if the note isn't
            "dirty"
        """
        # Cancel any queued save
        if self.__save_timeout_id is not None:
            GLib.source_remove(self.__save_timeout_id)
            self.__save_timeout_id = None

        # Persist any changes
        update = note.dirty or force_update_excerpt
        if update and not note.handling_conflict and (allow_empty or not note.new_and_empty):
            note.regenerate_excerpt()
            new_note = not note.has_id
            self.__persist_note_to_db(note)
            if new_note:
                self.add_notes_to_model([note])
        elif note.new_and_empty:
            self.__category_manager.note_deleted(note.category)
        else:
            self.apply_any_server_sanitised_title(note)
        note.handling_conflict = False

    def persist_note_category(self, note: Note, old_category: str) -> None:
        """Persist note category to database.

        :param Note note: The note
        :param str old_category: The old category
        """

        def remote_sync_and_integrate_to_models(is_new: bool) -> None:
            if not is_new:
                self.emit("sync-requested")

        def thread_do() -> None:
            is_new = not note.has_id
            if not is_new:
                self.__db.persist_note_category(note)
            GLib.idle_add(remote_sync_and_integrate_to_models, is_new)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()
        self.__category_manager.note_category_changed(old_category, note.category)

    def set_and_persist_favourite_for_notes(self, notes: list[Note], value: bool) -> bool:
        """Set and persist the favourite flag on the provided notes.

        :param list[Note] notes: The notes
        :param bool value: The new value
        :return: Whether any notes were updated
        :rtype: bool
        """
        updates = False
        for note in notes:
            if note.favourite == value:
                continue
            updates = True
            note.favourite = value
            note.flag_changed(update_last_modified=False)
            self.__db.persist_note_favourite(note)
        return updates

    def search_notes(self, term: str, restrict_to: list[int] = [], sort: bool = False) -> list[int]:
        """Search database for notes with provided term.

        :param str term: The search term
        :param list[int] restrict_to: List of ids of notes to restrict to
        :param bool sort: Whether the results should be datetime sorted
        :return: A list of ids for any matching notes
        :rtype: list[int]
        """
        return self.__db.search_notes(term, restrict_to, sort)

    def get_filtered_note_count(self, include_older_notes: bool) -> int:
        """Fetch filtered note count.

        :param bool include_older_notes: Whether to include the "older notes" model
        :return: The count
        :rtype: int
        """
        if include_older_notes:
            return len(self.__category_filtered_model)
        else:
            count = 0
            for model in (
                self.__favourites_model,
                self.__today_model,
                self.__yesterday_model,
                self.__week_model,
                self.__month_model,
                self.__last_month_model,
            ):
                count += len(model)
            return count

    def update_filters(
        self, category: Category, older_notes_displayed: bool, check_older_notes: bool
    ) -> bool:
        """Update model filters.

        If told to check whether older notes should be loaded the total number of notes in the
        category and within the last two months are compared to thresholds to determine whether to
        show the section automatically or not.

        :param Category category: Current category
        :param bool older_notes_displayed: Whether the notes older than two months section has
            been displayed
        :param bool check_older_notes: Whether to check if notes older than two months should be
            loaded
        :return: Whether the filter was opened to notes older than two months
        :rtype: bool
        """
        self.__category_filtered_model.invalidate_filter(category)
        self.__favourites_model.invalidate_filter()

        now = get_now()
        today_dt = now.replace(hour=0, minute=0, second=0, microsecond=0)
        today = today_dt.timestamp()
        yesterday = (today_dt - timedelta(days=1)).timestamp()
        monday = (today_dt - timedelta(days=now.weekday())).timestamp()
        start_of_month = now.replace(hour=0, minute=0, second=0, day=1).timestamp()
        start_of_last_month = get_start_of_last_month()

        self.__today_model.invalidate_filter(time_min=today)
        self.__yesterday_model.invalidate_filter(time_min=yesterday, time_max=today)
        self.__week_model.invalidate_filter(time_min=monday, time_max=yesterday)

        # If today is the first day of the month, or if the start of the week is earlier than the
        # start of the month don't show the rest of this month. Also ensures last month doesn't
        # have overlap content with the week.
        if today == start_of_month or monday < start_of_month:
            start_of_month = monday
        # Account for yesterday when setting rest of month end point
        month_end = yesterday if yesterday < monday else monday
        self.__month_model.invalidate_filter(time_min=start_of_month, time_max=month_end)

        # Account for yesterday when setting last month end point
        last_month_end = yesterday if yesterday < start_of_month else start_of_month
        self.__last_month_model.invalidate_filter(
            time_min=start_of_last_month, time_max=last_month_end
        )

        older_loaded = False
        if check_older_notes and not older_notes_displayed and self.__should_load_older_notes():
            if self.__older_notes_model is None:
                self.initiate_older_notes_model()
            older_loaded = True

        if older_notes_displayed or older_loaded:
            assert self.__older_notes_model is not None
            self.__older_notes_model.invalidate_filter(time_max=start_of_last_month)

        return older_loaded

    def filter_older_notes_by_date(self) -> None:
        """Filter older notes model to notes older than two months."""
        if self.__older_notes_model is None:
            return

        start_of_last_month = get_start_of_last_month()
        self.__older_notes_model.invalidate_filter(time_max=start_of_last_month)

    def get_deleted_notes_pending_undo(self) -> list[Note]:
        """Fetch any notes queued for deletion.

        :return: The notes
        :rtype: list[Note]
        """
        return self.__deleted_notes

    def set_deleted_notes_pending_undo(self, notes: list[Note]) -> None:
        """Set the notes queued for deletion.

        :param list[Note] notes: The notes
        """
        self.__deleted_notes = notes

    def apply_any_server_sanitised_title(self, note: Note) -> None:
        """Persist to the database any sanitised title received from the server while editing.

        During note creation we ignore sanitised titles returning from the server (to avoid
        replacing a title which is being edited via the first buffer line with something the
        server returns).

        :param Note note: The note to work on
        """
        if not note or note.server_sanitised_title is None:
            return

        logging.info("Applying sanitised title provided by server while creating note")
        note.title = note.server_sanitised_title
        note.server_sanitised_title = None
        changed_fields = DirtyFields()
        changed_fields.title = True
        self.__db.persist_note_selective(note, changed_fields)

    def get_note_failed_pushes(self, note: Note) -> list[FailedPush]:
        """Fetch any failed pushes for a note.

        :return: List of failed pushes
        :rtype: list[FailedPush]
        """
        pushes_as_str = self.__db.get_note_failed_pushes(note)
        if not pushes_as_str:
            return []

        try:
            pushes = json.loads(pushes_as_str)
        except json.JSONDecodeError as e:
            logging.warning(f"Failed to load failed pushes: {e}")
            return []

        if type(pushes) is not list:
            logging.warning("Failed to load failed pushes: not a list")
            return []

        ret = []
        for push_json in pushes:
            if not set(push_json.keys()) == {"hash", "length", "timestamp"}:
                logging.warning(f"Invalid push: {push_json}")
                continue
            ret.append(FailedPush(push_json["hash"], push_json["length"], push_json["timestamp"]))

        return ret

    def add_failed_push(self, note: Note, push: FailedPush) -> None:
        """Add a failed push to the database.

        :param Note note: The note to add for
        :param FailedPush push: The failed push info
        """
        pushes = self.get_note_failed_pushes(note)
        pushes.append(push)
        out = []
        for push in pushes:
            out.append(push._asdict())
        pushes_json = json.dumps(out)
        self.__db.persist_note_failed_pushes(note, pushes_json)

    @GObject.Property(type=NoteListModelBase, default=None)
    def base_model(self) -> NoteListModelBase:
        return self.__base_model

    @GObject.Property(type=NoteListModelFavourites, default=None)
    def favourites_model(self) -> NoteListModelFavourites:
        return self.__favourites_model

    @GObject.Property(type=NoteListModelTimeFiltered, default=None)
    def today_model(self) -> NoteListModelTimeFiltered:
        return self.__today_model

    @GObject.Property(type=NoteListModelTimeFiltered, default=None)
    def yesterday_model(self) -> NoteListModelTimeFiltered:
        return self.__yesterday_model

    @GObject.Property(type=NoteListModelTimeFiltered, default=None)
    def week_model(self) -> NoteListModelTimeFiltered:
        return self.__week_model

    @GObject.Property(type=NoteListModelTimeFiltered, default=None)
    def month_model(self) -> NoteListModelTimeFiltered:
        return self.__month_model

    @GObject.Property(type=NoteListModelTimeFiltered, default=None)
    def last_month_model(self) -> NoteListModelTimeFiltered:
        return self.__last_month_model

    @GObject.Property(type=NoteListModelTimeFiltered, default=None)
    def older_notes_model(self) -> Optional[NoteListModelTimeFiltered]:
        return self.__older_notes_model

    @GObject.Property(type=NoteListModelSearch, default=None)
    def search_model(self) -> Optional[NoteListModelSearch]:
        return self.__search_model

    def __persist_note_on_thread(self, note: Note, update_excerpt: bool) -> None:
        # Updated here (instead of in Note content setter) to reduce occurrence
        if update_excerpt:
            note.regenerate_excerpt()

        def remote_sync_and_integrate_new_note(note: Note, is_new: bool) -> None:
            if is_new:
                self.add_notes_to_model([note])
                self.emit("new-note-persisted")
            self.emit("sync-requested")

        def thread_do() -> None:
            is_new = not note.has_id
            self.__persist_note_to_db(note)

            GLib.idle_add(remote_sync_and_integrate_new_note, note, is_new)
            self.__save_timeout_id = None

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __should_load_older_notes(self) -> bool:
        total_note_count = len(self.__category_filtered_model)
        last_two_month_count = self.get_filtered_note_count(include_older_notes=False)

        if total_note_count == last_two_month_count:
            return True
        elif len(self.__category_filtered_model) < self.LOAD_ALL_NOTES_TOTAL_THRESHOLD:
            return True
        elif last_two_month_count < self.LOAD_ALL_NOTES_LAST_TWO_MONTHS_THRESHOLD:
            return True
        else:
            return False

    def __persist_note_to_db(self, note: Note) -> None:
        if note.id == -1:
            self.__db.add_note(note)
        else:
            self.__db.persist_note_editable(note)


def get_start_of_last_month() -> float:
    now = get_now()
    start_of_month_dt = now.replace(hour=0, minute=0, second=0, day=1)
    if start_of_month_dt.month > 1:
        start_of_last_month = start_of_month_dt.replace(
            month=start_of_month_dt.month - 1
        ).timestamp()
    else:
        start_of_last_month = start_of_month_dt.replace(
            month=12, year=start_of_month_dt.year - 1
        ).timestamp()
    return start_of_last_month


def get_now() -> datetime:
    return datetime.now()
